#!/usr/bin/env python
# coding: utf-8
from __future__ import print_function

__author__ = "Guillaume"
__license__ = "MIT"
__copyright__ = "2015, ESRF"
__date__ = "25/04/2020"

import unittest
from . import test_model
from . import test_align
from . import test_distance
from . import test_cormap
from . import test_autorg
from . import test_bift
from . import test_sasio
from . import test_sas_argparser
from . import test_fitting


def suite():
    testSuite = unittest.TestSuite()
    testSuite.addTest(test_bift.suite())
    testSuite.addTest(test_model.suite())
    testSuite.addTest(test_align.suite())
    testSuite.addTest(test_distance.suite())
    testSuite.addTest(test_cormap.suite())
    testSuite.addTest(test_autorg.suite())
    testSuite.addTest(test_sasio.suite())
    testSuite.addTest(test_sas_argparser.suite())
    testSuite.addTest(test_fitting.suite())
    return testSuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



# -*- coding: utf-8 -*-
#
#    Project: freesas
#             https://github.com/kif/freesas
#
#    Copyright (C) 2020  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
This module is mainly about the calculation of the Rambo-Tainer invariant
described in:

https://dx.doi.org/10.1038%2Fnature12070  

Some formula taken from Putnam et al, 2007, Table 1 in the review
"""
__authors__ = ["Martha E. Brennich", "J. Kieffer"]
__license__ = "MIT"
__date__ = "10/06/2020"

import logging
logger = logging.getLogger(__name__)
import numpy
from .collections import RT_RESULT


def extrapolate(data, guinier):
    """Extrapolate SAS data according to the Guinier fit until q=0
    Uncertainties are extrapolated (linearly) from the Guinier region 
    
    :param data: SAS data in q,I,dI format
    :param guinier: result of a Guinier fit
    :return: extrapolated SAS data 
    """
    
    dq = data[1, 0] - data[0, 0]
    qmin = data[guinier.start_point, 0]

    q_low = numpy.arange(0, qmin, dq)
    # Extrapolate I from Guinier approximation: 
    I_low = guinier.I0 * numpy.exp(-(q_low**2 * guinier.Rg**2) / 3.0)
    # Extrapolate dI from Guinier region:
    range_ = slice(guinier.start_point, guinier.end_point+1)
    slope, intercept = numpy.polyfit(data[range_, 0], data[range_, 2], deg=1)
    dI_low = abs(q_low*slope + intercept)
    # Now wrap-up
    data_low = numpy.vstack((q_low, I_low, dI_low)).T
    return numpy.concatenate((data_low, data[guinier.start_point:]))


def calc_Porod(data, guinier):
    """Calculate the particle volume according to Porod's formula:
    
    V = 2*π²I₀²/(sum_q I(q)q² dq)
    
    Formula from Putnam's review, 2007, table 1
    Intensities are extrapolated to q=0 using Guinier fit.
    
    :param data:  SAS data in q,I,dI format
    :param Guinier: result of a Guinier fit (instance of RT_RESULT)
    :return: Volume calculated according to Porrod's formula
    """ 
    q, I, dI = extrapolate(data, guinier).T
    
    denom = numpy.trapz(I*q**2, q)
    volume = 2*numpy.pi**2*guinier.I0 / denom
    return volume


def calc_Vc(data, Rg, dRg, I0, dI0, imin):
    """Calculates the Rambo-Tainer invariant Vc, including extrapolation to q=0
    
    :param data:  SAS data in q,I,dI format, cropped to maximal q that should be used for calculation (normally 2 nm-1)
    :param Rg,dRg,I0,dI0:  results from Guinier approximation/autorg
    :param imin:  minimal index of the Guinier range, below that index data will be extrapolated by the Guinier approximation
    :returns: Vc and an error estimate based on non-correlated error propagation
    """
    dq = data[1, 0] - data[0, 0]
    qmin = data[imin, 0]
    qlow = numpy.arange(0, qmin, dq)

    lowqint = numpy.trapz((qlow * I0 * numpy.exp(-(qlow * qlow * Rg * Rg) / 3.0)), qlow)
    dlowqint = numpy.trapz(qlow * numpy.sqrt((numpy.exp(-(qlow * qlow * Rg * Rg) / 3.0) * dI0) ** 2 + ((I0 * 2.0 * (qlow * qlow) * Rg / 3.0) * numpy.exp(-(qlow * qlow * Rg * Rg) / 3.0) * dRg) ** 2), qlow)
    vabs = numpy.trapz(data[imin:, 0] * data[imin:, 1], data[imin:, 0])
    dvabs = numpy.trapz(data[imin:, 0] * data[imin:, 2], data[imin:, 0])
    vc = I0 / (lowqint + vabs)
    dvc = (dI0 / I0 + (dlowqint + dvabs) / (lowqint + vabs)) * vc
    return (vc, dvc)


def calc_Rambo_Tainer(data,
                      guinier, qmax=2.0):
    """calculates the invariants Vc and Qr from the Rambo & Tainer 2013 Paper,
    also the the mass estimate based on Qr for proteins
    
    :param data: data in q,I,dI format, q in nm^-1
    :param guinier: RG_RESULT instance with result from the Guinier fit
    :param qmax: maximum q-value for the calculation in nm^-1
    @return: dict with Vc, Qr and mass plus errors
    """
    scale_prot = 1.0 / 0.1231
    power_prot = 1.0

    imax = abs(data[:, 0] - qmax).argmin()
    if (imax <= guinier.start_point) or (guinier.start_point < 0):  # unlikely but can happened
        logger.error("Guinier region start too late for Rambo_Tainer invariants calculation")
        return None
    vc = calc_Vc(data[:imax, :], guinier.Rg, guinier.sigma_Rg, guinier.I0, guinier.sigma_I0, guinier.start_point)

    qr = vc[0] ** 2 / (guinier.Rg)
    mass = scale_prot * qr ** power_prot

    dqr = qr * (guinier.sigma_Rg / guinier.Rg + 2 * ((vc[1]) / (vc[0])))
    dmass = mass * dqr / qr

    return RT_RESULT(vc[0], vc[1], qr, dqr, mass, dmass)



"""
This is the Python 3.8 implementation of mock_open taken from
https://github.com/python/cpython/blob/3.8/Lib/unittest/mock.py
Hence: 
"Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010,
2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020 Python Software Foundation;
All Rights Reserved"
"""

import io
from unittest.mock import MagicMock, DEFAULT


file_spec = None
#sentinel = _Sentinel()
#DEFAULT = sentinel.DEFAULT

def _to_stream(read_data):
    if isinstance(read_data, bytes):
        return io.BytesIO(read_data)
    else:
        return io.StringIO(read_data)


def mock_open(mock=None, read_data=''):
    """
    A helper function to create a mock to replace the use of `open`. It works
    for `open` called directly or used as a context manager.
    The `mock` argument is the mock object to configure. If `None` (the
    default) then a `MagicMock` will be created for you, with the API limited
    to methods or attributes available on standard file handles.
    `read_data` is a string for the `read`, `readline` and `readlines` of the
    file handle to return.  This is an empty string by default.
    """
    _read_data = _to_stream(read_data)
    _state = [_read_data, None]

    def _readlines_side_effect(*args, **kwargs):
        if handle.readlines.return_value is not None:
            return handle.readlines.return_value
        return _state[0].readlines(*args, **kwargs)

    def _read_side_effect(*args, **kwargs):
        if handle.read.return_value is not None:
            return handle.read.return_value
        return _state[0].read(*args, **kwargs)

    def _readline_side_effect(*args, **kwargs):
        yield from _iter_side_effect()
        while True:
            yield _state[0].readline(*args, **kwargs)

    def _iter_side_effect():
        if handle.readline.return_value is not None:
            while True:
                yield handle.readline.return_value
        for line in _state[0]:
            yield line

    def _next_side_effect():
        if handle.readline.return_value is not None:
            return handle.readline.return_value
        return next(_state[0])

    global file_spec
    if file_spec is None:
        import _io
        file_spec = list(set(dir(_io.TextIOWrapper)).union(set(dir(_io.BytesIO))))

    if mock is None:
        mock = MagicMock(name='open', spec=open)

    handle = MagicMock(spec=file_spec)
    handle.__enter__.return_value = handle

    handle.write.return_value = None
    handle.read.return_value = None
    handle.readline.return_value = None
    handle.readlines.return_value = None

    handle.read.side_effect = _read_side_effect
    _state[1] = _readline_side_effect()
    handle.readline.side_effect = _state[1]
    handle.readlines.side_effect = _readlines_side_effect
    handle.__iter__.side_effect = _iter_side_effect
    handle.__next__.side_effect = _next_side_effect

    def reset_data(*args, **kwargs):
        _state[0] = _to_stream(read_data)
        if handle.readline.side_effect == _state[1]:
            # Only reset the side effect if the user hasn't overridden it.
            _state[1] = _readline_side_effect()
            handle.readline.side_effect = _state[1]
        return DEFAULT

    mock.side_effect = reset_data
    mock.return_value = handle
    return mock



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2015-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
The silx package contains the following main sub-packages:

- silx.gui: Qt widgets for data visualization and data file browsing
- silx.image: Some processing functions for 2D images
- silx.io: Reading and writing data files (HDF5/NeXus, SPEC, ...)
- silx.math: Some processing functions for 1D, 2D, 3D, nD arrays
- silx.opencl: OpenCL-based data processing
- silx.sx: High-level silx functions suited for (I)Python console.
- silx.utils: Miscellaneous convenient functions

See silx documentation: http://www.silx.org/doc/silx/latest/
"""

__authors__ = ["Jérôme Kieffer"]
__license__ = "MIT"
__date__ = "31/08/2018"

import os as _os
import logging as _logging

_logging.getLogger(__name__).addHandler(_logging.NullHandler())


project = _os.path.basename(_os.path.dirname(_os.path.abspath(__file__)))

try:
    from ._version import __date__ as date  # noqa
    from ._version import (
        version,
        version_info,
        hexversion,
        strictversion,
        dated_version,
    )  # noqa
except ImportError:
    raise RuntimeError(
        "Do NOT use %s from its sources: build it and use the built version"
        % project
    )



#!/usr/bin/env python
# coding: utf-8
from __future__ import print_function

__author__ = "Guillaume"
__license__ = "MIT"
__copyright__ = "2015, ESRF"

import os
from math import sqrt
import threading
import six
import numpy
try:
    from . import _distance
except ImportError:
    _distance = None
from . import transformations


def delta_expand(vec1, vec2):
    """Create a 2d array with the difference vec1[i]-vec2[j]

    :param vec1, vec2: 1d-array
    :return v1 - v2: difference for any element of v1 and v2 (i.e a 2D array)
    """
    v1 = numpy.ascontiguousarray(vec1)
    v2 = numpy.ascontiguousarray(vec2)
    v1.shape = -1, 1
    v2.shape = 1, -1
    v1.strides = v1.strides[0], 0
    v2.strides = 0, v2.strides[-1]
    return v1 - v2


class SASModel:
    """
    Tools for Dummy Atoms Model manipulation
    """

    def __init__(self, molecule=None):
        """
        :param molecule: if str, name of a pdb file, else if 2d-array, coordinates of atoms of a molecule
        """
        if isinstance(molecule, (six.text_type, six.binary_type)) and os.path.exists(molecule):
            self.read(molecule)
        else:
            self.atoms = molecule if molecule is not None else []  # initial coordinates of each dummy atoms of the molecule, fourth column full of one for the transformation matrix
            self.header = ""  # header of the PDB file
            self.rfactor = None
        self.radius = 1.0  # unused at the moment
        self.com = []
        self._fineness = None
        self._Rg = None
        self._Dmax = None
        self.inertensor = []
        self.can_param = []
        self.enantiomer = None  # symmetry used on the molecule
        self._sem = threading.Semaphore()

    def __repr__(self):
        return "SAS model with %i atoms" % len(self.atoms)

    def read(self, filename):
        """
        Read the PDB file,
        extract coordinates of each dummy atom,
        extract the R-factor of the model, coordinates of each dummy atom and pdb file header.

        :param filename: name of the pdb file to read
        """
        header = []
        atoms = []
        with open(filename) as fd:
            for line in fd:
                if line.startswith("ATOM"):
                    x = float(line[30:38])
                    y = float(line[38:46])
                    z = float(line[46:54])
                    atoms.append([x, y, z])
                if line.startswith("REMARK 265  Final R-factor"):  # very dependent of the pdb file format !
                    self.rfactor = float(line[43:56])
                header.append(line)
        self.header = header
        atom3 = numpy.array(atoms)
        self.atoms = numpy.append(atom3, numpy.ones((atom3.shape[0], 1), dtype="float"), axis=1)

    def save(self, filename):
        """
        Save the position of each dummy atom in a PDB file.

        :param filename: name of the pdb file to write
        """
        nr = 0
        self.atoms = numpy.delete(self.atoms, 3, 1)
        with open(filename, "w") as pdbout:
            for line in self.header:
                if line.startswith("ATOM"):
                    if nr < self.atoms.shape[0]:
                        line = line[:30] + "%8.3f%8.3f%8.3f" % tuple(self.atoms[nr]) + line[54:]
                    else:
                        line = ""
                    nr += 1
                pdbout.write(line)

    def centroid(self):
        """
        Calculate the position of the center of mass of the molecule.

        :return self.com: 1d array, coordinates of the center of mass of the molecule
        """
        mol = self.atoms[:, 0:3]
        self.com = mol.mean(axis=0)
        return self.com

    def inertiatensor(self):
        """
        calculate the inertia tensor of the protein

        :return self.inertensor: inertia tensor of the molecule
        """
        if len(self.com) == 0:
            self.com = self.centroid()

        mol = self.atoms[:, 0:3] - self.com
        self.inertensor = numpy.empty((3, 3), dtype="float")
        delta_kron = lambda i, j: 1 if i == j else 0
        for i in range(3):
            for j in range(i, 3):
                self.inertensor[i, j] = self.inertensor[j, i] = (delta_kron(i, j) * (mol ** 2).sum(axis=1) - (mol[:, i] * mol[:, j])).sum() / mol.shape[0]
        return self.inertensor

    def canonical_translate(self):
        """
        Calculate the translation matrix to translate the center of mass of the molecule on the origin of the base.

        :return trans: translation matrix
        """
        if len(self.com) == 0:
            self.com = self.centroid()

        trans = numpy.identity(4, dtype="float")
        trans[0:3, 3] = -self.com
        return trans

    def canonical_rotate(self):
        """
        Calculate the rotation matrix to align inertia momentum of the molecule on principal axis.

        :return rot: rotation matrix det==1
        """
        if len(self.inertensor) == 0:
            self.inertensor = self.inertiatensor()

        w, v = numpy.linalg.eigh(self.inertensor)
        mat = v[:, w.argsort()]

        rot = numpy.zeros((4, 4), dtype="float")
        rot[3, 3] = 1
        rot[:3, :3] = mat.T

        det = numpy.linalg.det(mat)
        if det > 0:
            self.enantiomer = [1, 1, 1]
        else:
            self.enantiomer = [-1, -1, -1]
            mirror = numpy.array([[-1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]], dtype="float")
            rot = numpy.dot(mirror, rot)

        return rot

    def canonical_parameters(self):
        """
        Save the 6 canonical parameters of the initial molecule:
        x0, y0, z0, the position of the center of mass
        phi, theta, psi, the three Euler angles of the canonical rotation (axis:x,y',z'')
        """
        rot = self.canonical_rotate()
        trans = self.canonical_translate()

        angles = transformations.euler_from_matrix(rot)
        shift = transformations.translation_from_matrix(trans)
        self.can_param = [shift[0], shift[1], shift[2], angles[0], angles[1], angles[2]]

    def calc_invariants(self, use_cython=True):
        """
        Calculate the invariants of the structure:
            * fineness, ie. average distance between an atoms and its nearest neighbor
            * radius of gyration of the model
            * diameter of the model

        :return invariants: 3-tuple containing (fineness, Rg, Dmax)
        """
        if _distance and use_cython:
            return _distance.calc_invariants(self.atoms)

        else:
            size = self.atoms.shape[0]
            D = delta_expand(self.atoms[:, 0], self.atoms[:, 0]) ** 2 + delta_expand(self.atoms[:, 1], self.atoms[:, 1]) ** 2 + delta_expand(self.atoms[:, 2], self.atoms[:, 2]) ** 2
            Rg = sqrt(D.sum() / 2.0) / size
            Dmax = sqrt(D.max())
            d12 = (D.max() * numpy.eye(size) + D).min(axis=0).mean()
            fineness = sqrt(d12)
            return fineness, Rg, Dmax

    @property
    def fineness(self):
        if self._fineness is None:
            with self._sem:
                if self._fineness is None:
                    self._fineness, self._Rg, self._Dmax = self.calc_invariants()
        return self._fineness

    @property
    def Rg(self):
        if self._Rg is None:
            with self._sem:
                if self._Rg is None:
                    self._fineness, self._Rg, self._Dmax = self.calc_invariants()
        return self._Rg

    @property
    def Dmax(self):
        if self._Dmax is None:
            with self._sem:
                if self._Dmax is None:
                    self._fineness, self._Rg, self._Dmax = self.calc_invariants()
        return self._Dmax

    def dist(self, other, molecule1, molecule2, use_cython=True):
        """
        Calculate the distance with another model.

        :param self,other: two SASModel
        :param molecule1: 2d array of the position of each atom of the first molecule
        :param molecule2: 2d array of the position of each atom of the second molecule
        :return D: NSD between the 2 molecules, in their position molecule1 and molecule2
        """
        if _distance and use_cython:
            return _distance.calc_distance(molecule1, molecule2, self.fineness, other.fineness)

        else:
            mol1 = molecule1[:, 0:3]
            mol2 = molecule2[:, 0:3]

            mol1x = mol1[:, 0]
            mol1y = mol1[:, 1]
            mol1z = mol1[:, 2]
            mol1x.shape = mol1.shape[0], 1
            mol1y.shape = mol1.shape[0], 1
            mol1z.shape = mol1.shape[0], 1

            mol2x = mol2[:, 0]
            mol2y = mol2[:, 1]
            mol2z = mol2[:, 2]
            mol2x.shape = mol2.shape[0], 1
            mol2y.shape = mol2.shape[0], 1
            mol2z.shape = mol2.shape[0], 1

            d2 = delta_expand(mol1x, mol2x) ** 2 + delta_expand(mol1y, mol2y) ** 2 + delta_expand(mol1z, mol2z) ** 2

            D = (0.5 * ((1. / ((mol1.shape[0]) * other.fineness * other.fineness)) * (d2.min(axis=1).sum()) + (1. / ((mol2.shape[0]) * self.fineness * self.fineness)) * (d2.min(axis=0)).sum())) ** 0.5
            return D

    def transform(self, param, symmetry, reverse=None):
        """
        Calculate the new coordinates of each dummy atoms of the molecule after a transformation defined by six parameters and a symmetry

        :param param: 6 parameters of transformation (3 coordinates of translation, 3 Euler angles)
        :param symmetry: list of three constants which define a symmetry to apply
        :return mol: 2d array, coordinates after transformation
        """
        mol = self.atoms

        sym = numpy.array([[symmetry[0], 0, 0, 0], [0, symmetry[1], 0, 0], [0, 0, symmetry[2], 0], [0, 0, 0, 1]], dtype="float")
        if not reverse:
            vect = numpy.array([param[0:3]])
            angles = (param[3:6])

            translat1 = transformations.translation_matrix(vect)
            rotation = transformations.euler_matrix(*angles)
            translat2 = numpy.dot(numpy.dot(rotation, translat1), rotation.T)
            transformation = numpy.dot(translat2, rotation)

        else:
            vect = -numpy.array([param[0:3]])
            angles = (-param[5], -param[4], -param[3])

            translat = transformations.translation_matrix(vect)
            rotation = transformations.euler_matrix(*angles, axes="szyx")
            transformation = numpy.dot(translat, rotation)

        mol = numpy.dot(transformation, mol.T)
        mol = numpy.dot(sym, mol).T
        return mol

    def dist_after_movement(self, param, other, symmetry):
        """
        The first molecule, molref, is put on its canonical position.
        The second one, mol2, is moved following the transformation selected

        :param param: list of 6 parameters for the transformation, 3 coordinates of translation and 3 Euler angles
        :param symmetry: list of three constants which define a symmetry to apply
        :return distance: the NSD between the first molecule and the second one after its movement
        """
        if not self.can_param:
            self.canonical_parameters()

        can_param1 = self.can_param
        molref_can = self.transform(can_param1, [1, 1, 1])  # molecule reference put on its canonical position

        mol2_moved = other.transform(param, symmetry)  # movement selected applied to mol2
        distance = self.dist(other, molref_can, mol2_moved)

        return distance



#!/usr/bin/python
# coding: utf-8
from __future__ import print_function

__author__ = "Guillaume"
__license__ = "MIT"
__copyright__ = "2015, ESRF"

import numpy
import unittest
from .utilstests import get_datafile
from ..model import SASModel
from ..average import Grid, AverModels

import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("AlignModels_test")


class TestAverage(unittest.TestCase):
    testfile1 = get_datafile("model-01.pdb")
    testfile2 = get_datafile("model-02.pdb")
    inputfiles = [testfile1, testfile2]
    grid = Grid(inputfiles)

    def test_gridsize(self):
        inputfiles = self.inputfiles
        grid = self.grid
        size = grid.spatial_extent()
        coordmax = numpy.array([size[0:3]], dtype="float")
        coordmin = numpy.array([size[3:6]], dtype="float")

        pb = 0
        for i in inputfiles:
            m = SASModel(i)
            a = coordmin + m.atoms[:, 0:3]
            b = m.atoms[:, 0:3] - coordmax
            if (a >= 0.0).any() or (b >= 0.0).any():
                pb += 1
        self.assertEqual(pb, 0, msg="computed size is not the good one")

    def test_knots(self):
        grid = self.grid
        nbknots = numpy.random.randint(4000, 6000)
        threshold = 10.0  # acceptable difference between nbknots and the effective nb of knots in percent
        grid.calc_radius(nbknots)
        grid.make_grid()
        gap = (1.0 * (grid.nbknots - nbknots) / nbknots) * 100
        self.assertGreater(threshold, gap, msg="final number of knots too different of wanted number: %s != %s" % (nbknots, grid.nbknots))

    def test_makegrid(self):
        grid = self.grid
        lattice = grid.make_grid()
        m = SASModel(lattice)
        self.assertAlmostEqual(m.fineness, 2 * grid.radius, 10, msg="grid do not have the computed radius")

    def test_read(self):
        inputfiles = self.inputfiles
        average = AverModels(inputfiles, self.grid.coordknots)
        models = [SASModel(inputfiles[1]), SASModel(inputfiles[0])]
        average.read_files(reference=1)
        diff = 0.0
        for i in range(len(inputfiles)):
            diff += (models[i].atoms - average.models[i].atoms).max()
        self.assertAlmostEqual(diff, 0.0, 10, msg="Files not read properly")

    def test_occupancy(self):
        average = AverModels(self.inputfiles, self.grid.coordknots)
        average.read_files()
        occ_grid = average.assign_occupancy()
        average.grid = occ_grid
        assert occ_grid.shape[-1] == 5, "problem in grid shape"
        diff = occ_grid[:-1, 3] - occ_grid[1:, 3]
        self.assertTrue(diff.max() >= 0.0, msg="grid is not properly sorted with occupancy")


def suite():
    testSuite = unittest.TestSuite()
    testSuite.addTest(TestAverage("test_gridsize"))
    testSuite.addTest(TestAverage("test_knots"))
    testSuite.addTest(TestAverage("test_makegrid"))
    testSuite.addTest(TestAverage("test_read"))
    testSuite.addTest(TestAverage("test_occupancy"))
    return testSuite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python3
# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2015-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Run the tests of the project.

This script expects a suite function in <project_package>.test,
which returns a unittest.TestSuite.

Test coverage dependencies: coverage, lxml.
"""

__authors__ = ["Jérôme Kieffer", "Thomas Vincent"]
__date__ = "16/04/2020"
__license__ = "MIT"

import distutils.util
import logging
import os
import subprocess
import sys
import time
import unittest
import collections
from argparse import ArgumentParser


class StreamHandlerUnittestReady(logging.StreamHandler):
    """The unittest class TestResult redefine sys.stdout/err to capture
    stdout/err from tests and to display them only when a test fail.
    This class allow to use unittest stdout-capture by using the last sys.stdout
    and not a cached one.
    """

    def emit(self, record):
        """
        :type record: logging.LogRecord
        """
        self.stream = sys.stderr
        super(StreamHandlerUnittestReady, self).emit(record)

    def flush(self):
        pass


def createBasicHandler():
    """Create the handler using the basic configuration"""
    hdlr = StreamHandlerUnittestReady()
    fs = logging.BASIC_FORMAT
    dfs = None
    fmt = logging.Formatter(fs, dfs)
    hdlr.setFormatter(fmt)
    return hdlr


# Use an handler compatible with unittests, else use_buffer is not working
logging.root.addHandler(createBasicHandler())

# Capture all default warnings
logging.captureWarnings(True)
import warnings

warnings.simplefilter("default")

logger = logging.getLogger("run_tests")
logger.setLevel(logging.WARNING)

logger.info("Python %s %s", sys.version, tuple.__itemsize__ * 8)


try:
    import resource
except ImportError:
    resource = None
    logger.warning("resource module missing")

import importlib

importer = importlib.import_module

try:
    import numpy
except Exception as error:
    logger.warning("Numpy missing: %s", error)
else:
    logger.info("Numpy %s", numpy.version.version)

try:
    import h5py
except Exception as error:
    logger.warning("h5py missing: %s", error)
else:
    logger.info("h5py %s", h5py.version.version)


def get_project_name(root_dir):
    """Retrieve project name by running python setup.py --name in root_dir.

    :param str root_dir: Directory where to run the command.
    :return: The name of the project stored in root_dir
    """
    logger.debug("Getting project name in %s", root_dir)
    p = subprocess.Popen(
        [sys.executable, "setup.py", "--name"],
        shell=False,
        cwd=root_dir,
        stdout=subprocess.PIPE,
    )
    name, _stderr_data = p.communicate()
    logger.debug("subprocess ended with rc= %s", p.returncode)
    return name.split()[-1].decode("ascii")


class TextTestResultWithSkipList(unittest.TextTestResult):
    """Override default TextTestResult to display list of skipped tests at the
    end
    """

    def printErrors(self):
        unittest.TextTestResult.printErrors(self)
        # Print skipped tests at the end
        self.printGroupedList("SKIPPED", self.skipped)

    def printGroupedList(self, flavour, errors):
        grouped = collections.OrderedDict()

        for test, err in errors:
            if err in grouped:
                grouped[err] = grouped[err] + [test]
            else:
                grouped[err] = [test]

        for err, tests in grouped.items():
            self.stream.writeln(self.separator1)
            for test in tests:
                self.stream.writeln(
                    "%s: %s" % (flavour, self.getDescription(test))
                )
            self.stream.writeln(self.separator2)
            self.stream.writeln("%s" % err)


class ProfileTextTestResult(unittest.TextTestRunner.resultclass):
    def __init__(self, *arg, **kwarg):
        unittest.TextTestRunner.resultclass.__init__(self, *arg, **kwarg)
        self.logger = logging.getLogger("memProf")
        self.logger.setLevel(min(logging.INFO, logging.root.level))
        self.logger.handlers.append(logging.FileHandler("profile.log"))

    def startTest(self, test):
        unittest.TextTestRunner.resultclass.startTest(self, test)
        if resource:
            self.__mem_start = resource.getrusage(
                resource.RUSAGE_SELF
            ).ru_maxrss
        self.__time_start = time.time()

    def stopTest(self, test):
        unittest.TextTestRunner.resultclass.stopTest(self, test)
        # see issue 311. For other platform, get size of ru_maxrss in "man getrusage"
        if sys.platform == "darwin":
            ratio = 1e-6
        else:
            ratio = 1e-3
        if resource:
            memusage = (
                resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
                - self.__mem_start
            ) * ratio
        else:
            memusage = 0
        self.logger.info(
            "Time: %.3fs \t RAM: %.3f Mb\t%s",
            time.time() - self.__time_start,
            memusage,
            test.id(),
        )


def report_rst(cov, package, version="0.0.0", base=""):
    """
    Generate a report of test coverage in RST (for Sphinx inclusion)

    :param cov: test coverage instance
    :param str package: Name of the package
    :param str base: base directory of modules to include in the report
    :return: RST string
    """
    import tempfile

    fd, fn = tempfile.mkstemp(suffix=".xml")
    os.close(fd)
    cov.xml_report(outfile=fn)

    from lxml import etree

    xml = etree.parse(fn)
    classes = xml.xpath("//class")

    line0 = "Test coverage report for %s" % package
    res = [line0, "=" * len(line0), ""]
    res.append(
        "Measured on *%s* version %s, %s"
        % (package, version, time.strftime("%d/%m/%Y"))
    )
    res += [
        "",
        ".. csv-table:: Test suite coverage",
        '   :header: "Name", "Stmts", "Exec", "Cover"',
        "   :widths: 35, 8, 8, 8",
        "",
    ]
    tot_sum_lines = 0
    tot_sum_hits = 0

    for cl in classes:
        name = cl.get("name")
        fname = cl.get("filename")
        if not os.path.abspath(fname).startswith(base):
            continue
        lines = cl.find("lines").getchildren()
        hits = [int(i.get("hits")) for i in lines]

        sum_hits = sum(hits)
        sum_lines = len(lines)

        cover = 100.0 * sum_hits / sum_lines if sum_lines else 0

        if base:
            name = os.path.relpath(fname, base)

        res.append(
            '   "%s", "%s", "%s", "%.1f %%"'
            % (name, sum_lines, sum_hits, cover)
        )
        tot_sum_lines += sum_lines
        tot_sum_hits += sum_hits
    res.append("")
    res.append(
        '   "%s total", "%s", "%s", "%.1f %%"'
        % (
            package,
            tot_sum_lines,
            tot_sum_hits,
            100.0 * tot_sum_hits / tot_sum_lines if tot_sum_lines else 0,
        )
    )
    res.append("")
    return os.linesep.join(res)


def is_debug_python():
    """Returns true if the Python interpreter is in debug mode."""
    import sysconfig

    if sysconfig.get_config_var("Py_DEBUG"):
        return True

    return hasattr(sys, "gettotalrefcount")


def build_project(name, root_dir):
    """Run python setup.py build for the project.

    Build directory can be modified by environment variables.

    :param str name: Name of the project.
    :param str root_dir: Root directory of the project
    :return: The path to the directory were build was performed
    """
    platform = distutils.util.get_platform()
    architecture = "lib.%s-%i.%i" % (
        platform,
        sys.version_info[0],
        sys.version_info[1],
    )
    if is_debug_python():
        architecture += "-pydebug"

    if os.environ.get("PYBUILD_NAME") == name:
        # we are in the debian packaging way
        home = os.environ.get("PYTHONPATH", "").split(os.pathsep)[-1]
    elif os.environ.get("BUILDPYTHONPATH"):
        home = os.path.abspath(os.environ.get("BUILDPYTHONPATH", ""))
    else:
        home = os.path.join(root_dir, "build", architecture)

    logger.warning("Building %s to %s", name, home)
    p = subprocess.Popen(
        [sys.executable, "setup.py", "build"], shell=False, cwd=root_dir
    )
    logger.debug("subprocess ended with rc= %s", p.wait())

    if os.path.isdir(home):
        return home
    alt_home = os.path.join(os.path.dirname(home), "lib")
    if os.path.isdir(alt_home):
        return alt_home


def import_project_module(project_name, project_dir):
    """Import project module, from the system of from the project directory"""
    # Prevent importing from source directory
    if os.path.dirname(os.path.abspath(__file__)) == os.path.abspath(
        sys.path[0]
    ):
        removed_from_sys_path = sys.path.pop(0)
        logger.info("Patched sys.path, removed: '%s'", removed_from_sys_path)

    if "--installed" in sys.argv:
        try:
            module = importer(project_name)
        except ImportError:
            raise ImportError(
                "%s not installed: Cannot run tests on installed version"
                % PROJECT_NAME
            )
    else:  # Use built source
        build_dir = build_project(project_name, project_dir)
        if build_dir is None:
            logging.error("Built project is not available !!! investigate")
        sys.path.insert(0, build_dir)
        logger.warning("Patched sys.path, added: '%s'", build_dir)
        module = importer(project_name)
    return module


def get_test_options(project_module):
    """Returns the test options if available, else None"""
    module_name = project_module.__name__ + ".test.utils"
    logger.info("Import %s", module_name)
    try:
        test_utils = importer(module_name)
    except ImportError:
        logger.warning(
            "No module named '%s'. No test options available.", module_name
        )
        return None

    test_options = getattr(test_utils, "test_options", None)
    return test_options


if __name__ == "__main__":  # Needed for multiprocessing support on Windows
    PROJECT_DIR = os.path.dirname(os.path.abspath(__file__))
    PROJECT_NAME = get_project_name(PROJECT_DIR)
    logger.info("Project name: %s", PROJECT_NAME)

    project_module = import_project_module(PROJECT_NAME, PROJECT_DIR)
    PROJECT_VERSION = getattr(project_module, "version", "")
    PROJECT_PATH = project_module.__path__[0]

    test_options = get_test_options(project_module)
    """Contains extra configuration for the tests."""

    epilog = """Environment variables:
    SILX_OPENCL=False to disable OpenCL tests
    SILX_TEST_LOW_MEM=True to disable tests taking large amount of memory
    GPU=False to disable the use of a GPU with OpenCL test
    WITH_GL_TEST=False to disable tests using OpenGL
    """
    parser = ArgumentParser(description="Run the tests.", epilog=epilog)

    parser.add_argument(
        "--installed",
        action="store_true",
        dest="installed",
        default=False,
        help=(
            "Test the installed version instead of"
            + "building from the source"
        ),
    )
    parser.add_argument(
        "-c",
        "--coverage",
        dest="coverage",
        action="store_true",
        default=False,
        help=(
            "Report code coverage" + "(requires 'coverage' and 'lxml' module)"
        ),
    )
    parser.add_argument(
        "-m",
        "--memprofile",
        dest="memprofile",
        action="store_true",
        default=False,
        help="Report memory profiling",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        default=0,
        action="count",
        dest="verbose",
        help="Increase verbosity. Option -v prints additional "
        + "INFO messages. Use -vv for full verbosity, "
        + "including debug messages and test help strings.",
    )
    if test_options is not None:
        test_options.add_parser_argument(parser)

    default_test_name = "%s.test.suite" % PROJECT_NAME
    parser.add_argument(
        "test_name",
        nargs="*",
        default=(default_test_name,),
        help="Test names to run (Default: %s)" % default_test_name,
    )
    options = parser.parse_args()
    sys.argv = [sys.argv[0]]

    test_verbosity = 1
    use_buffer = True
    if options.verbose == 1:
        logging.root.setLevel(logging.INFO)
        logger.info("Set log level: INFO")
        test_verbosity = 2
        use_buffer = False
    elif options.verbose > 1:
        logging.root.setLevel(logging.DEBUG)
        logger.info("Set log level: DEBUG")
        test_verbosity = 2
        use_buffer = False

    if options.coverage:
        logger.info("Running test-coverage")
        import coverage

        omits = [
            "*test*",
            "*third_party*",
            "*/setup.py",
            # temporary test modules (silx.math.fit.test.test_fitmanager)
            "*customfun.py",
        ]
        try:
            cov = coverage.Coverage(omit=omits)
        except AttributeError:
            cov = coverage.coverage(omit=omits)
        cov.start()

    # Run the tests
    runnerArgs = {}
    runnerArgs["verbosity"] = test_verbosity
    runnerArgs["buffer"] = use_buffer
    if options.memprofile:
        runnerArgs["resultclass"] = ProfileTextTestResult
    else:
        runnerArgs["resultclass"] = TextTestResultWithSkipList
    runner = unittest.TextTestRunner(**runnerArgs)

    logger.warning(
        "Test %s %s from %s", PROJECT_NAME, PROJECT_VERSION, PROJECT_PATH
    )

    test_module_name = PROJECT_NAME + ".test"
    logger.info("Import %s", test_module_name)
    test_module = importer(test_module_name)
    test_suite = unittest.TestSuite()

    if test_options is not None:
        # Configure the test options according to the command lines and the the environment
        test_options.configure(options)
    else:
        logger.warning("No test options available.")

    if not options.test_name:
        # Do not use test loader to avoid cryptic exception
        # when an error occur during import
        project_test_suite = getattr(test_module, "suite")
        test_suite.addTest(project_test_suite())
    else:
        test_suite.addTest(
            unittest.defaultTestLoader.loadTestsFromNames(options.test_name)
        )

    # Display the result when using CTRL-C
    unittest.installHandler()

    result = runner.run(test_suite)

    if result.wasSuccessful():
        exit_status = 0
    else:
        exit_status = 1

    if options.coverage:
        cov.stop()
        cov.save()
        with open("coverage.rst", "w") as fn:
            fn.write(
                report_rst(cov, PROJECT_NAME, PROJECT_VERSION, PROJECT_PATH)
            )

    sys.exit(exit_status)



#!/usr/bin/python3
# coding: utf-8
#
#    Project: freesas
#             https://github.com/kif/freesas
#
#    Copyright (C) 2020  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__author__ = ["Jérôme Kieffer", "Martha Brennich"]
__license__ = "MIT"
__copyright__ = "2021, ESRF"
__date__ = "19/03/2021"

import sys
import logging
from freesas.autorg import auto_guinier
from freesas.sas_argparser import GuinierParser
from freesas.fitting import run_guinier_fit

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger("auto_guinier")

if sys.version_info < (3, 6):
    logger.error("This code uses F-strings and requires Python 3.6+")


def build_parser() -> GuinierParser:
    """Build parser for input and return list of files.
    :return: parser
    """
    description = (
        "Calculate the radius of gyration using linear fitting of"
        "logarithmic intensities for a set of scattering curves"
    )
    epilog = """free_guinier is an open-source implementation of
    the autorg algorithm originately part of the ATSAS suite.
    As this tool used a different theory, some results may differ
    """
    return GuinierParser(
        prog="free_guinier", description=description, epilog=epilog
    )


def main() -> None:
    """Entry point for free_guinier app"""
    parser = build_parser()
    run_guinier_fit(fit_function=auto_guinier, parser=parser, logger=logger)


if __name__ == "__main__":
    main()



#!/usr/bin/python3
__author__ = "Guillaume Bonamis"
__license__ = "MIT"
__copyright__ = "2015, ESRF"
__date__ = "09/07/2020"

import logging
from os.path import dirname, abspath
from freesas.align import InputModels, AlignModels
from freesas.sas_argparser import SASParser

base = dirname(dirname(abspath(__file__)))

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("supycomb")


def parse():

    """Parse input and return list of files.
    :return: list of args
    """
    description = "Align several models and calculate NSD"
    epilog = """supycomb is an open-source implementation of
    [J. Appl. Cryst. (2001). 34, 33-41](doi:10.1107/S0021889800014126).

    The main difference with supcomb: the fast mode does not re-bin beads. It only refines the best matching orientation which provides a speed-up of a factor 8.

    """
    parser = SASParser(prog="supycomp", description=description, epilog=epilog)
    parser.add_file_argument(help_text="pdb files to align")
    parser.add_argument(
        "-m",
        "--mode",
        dest="mode",
        type=str,
        choices=["SLOW", "FAST"],
        default="SLOW",
        help="Either SLOW or FAST, default: %(default)s)",
    )
    parser.add_argument(
        "-e",
        "--enantiomorphs",
        type=str,
        choices=["YES", "NO"],
        default="YES",
        help="Search enantiomorphs, YES or NO, default: %(default)s)",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        type=str,
        choices=["ON", "OFF"],
        default="ON",
        help="Hide log or not, default: %(default)s",
    )
    parser.add_argument(
        "-g",
        "--gui",
        type=str,
        choices=["YES", "NO"],
        default="YES",
        help="Use GUI for figures or not, default: %(default)s",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default="aligned.pdb",
        help="output filename, default: %(default)s",
    )
    return parser.parse_args()


def main():
    """main application"""

    args = parse()
    input_len = len(args.file)
    logger.info("%s input files" % input_len)
    selection = InputModels()

    if args.mode == "SLOW":
        slow = True
        logger.info("SLOW mode")
    else:
        slow = False
        logger.info("FAST mode")

    if args.enantiomorphs == "YES":
        enantiomorphs = True
    else:
        enantiomorphs = False
        logger.info("NO enantiomorphs")

    if args.quiet == "OFF":
        logger.setLevel(logging.DEBUG)
        logger.info("setLevel: Debug")

    if args.gui == "NO":
        save = True
        logger.info(
            "Figures saved automatically : \n  R factor values and selection =>  Rfactor.png \n  NSD table and selection =>  nsd.png"
        )
    else:
        save = False

    align = AlignModels(args.file, slow=slow, enantiomorphs=enantiomorphs)
    if input_len == 2:
        align.outputfiles = args.output
        align.assign_models()
        dist = align.alignment_2models()
        logger.info("%s and %s aligned" % (args.file[0], args.file[1]))
        logger.info("NSD after optimized alignment = %.2f" % dist)
    else:
        align.outputfiles = [
            "model-%02i.pdb" % (i + 1) for i in range(input_len)
        ]
        selection.inputfiles = args.file
        selection.models_selection()
        selection.rfactorplot(save=save)
        align.models = selection.sasmodels
        align.validmodels = selection.validmodels

        align.makeNSDarray()
        align.alignment_reference()
        logger.info(
            "valid models aligned on the model %s" % (align.reference + 1)
        )
        align.plotNSDarray(rmax=round(selection.rmax, 4), save=save)

    if not save and input_len > 2:
        input("Press any key to exit")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
#    Project: freesas
#             https://github.com/kif/freesas
#
#    Copyright (C) 2020  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Set of namedtuples defined a bit everywhere
"""
__authors__ = ["Jérôme Kieffer"]
__license__ = "MIT"
__copyright__ = "2020 ESRF"
__date__ = "13/10/2020"

from collections import namedtuple
from os import linesep
import numpy

# Used in AutoRg
RG_RESULT = namedtuple(
    "RG_RESULT",
    "Rg sigma_Rg I0 sigma_I0 start_point end_point quality aggregated",
)


def _RG_RESULT_repr(self):
    return f"Rg={self.Rg:6.4f}(±{self.sigma_Rg:6.4f}) I0={self.I0:6.4f}(±{self.sigma_I0:6.4f}) [{self.start_point}-{self.end_point}] {100.0*self.quality:5.2f}% {'aggregated' if self.aggregated>0.1 else ''}"


RG_RESULT.__repr__ = _RG_RESULT_repr

FIT_RESULT = namedtuple(
    "FIT_RESULT",
    "slope sigma_slope intercept sigma_intercept, R, R2, chi2, RMSD",
)
RT_RESULT = namedtuple("RT_RESULT", "Vc sigma_Vc Qr sigma_Qr mass sigma_mass")


def _RT_RESULT_repr(self):
    return f"Vc={self.Vc:6.4f}(±{self.sigma_Vc:6.4f}) Qr={self.Qr:6.4f}(±{self.sigma_Qr:6.4f}) mass={self.mass:6.4f}(±{self.sigma_mass:6.4f})"


RT_RESULT.__repr__ = _RT_RESULT_repr

# Used in BIFT
RadiusKey = namedtuple("RadiusKey", "Dmax npt")
PriorKey = namedtuple("PriorKey", "type npt")
TransfoValue = namedtuple("TransfoValue", "transfo B sum_dia")
EvidenceKey = namedtuple("EvidenceKey", "Dmax alpha npt")
EvidenceResult = namedtuple(
    "EvidenceResult", "evidence chi2r regularization radius density converged"
)

StatsResult = namedtuple(
    "StatsResult",
    "radius density_avg density_std evidence_avg evidence_std Dmax_avg Dmax_std alpha_avg, alpha_std chi2r_avg chi2r_std regularization_avg regularization_std Rg_avg Rg_std I0_avg I0_std",
)


def save_bift(stats, filename, source=None):
    "Save the results of the fit to the file"
    res = [
        "Dmax= %.2f±%.2f" % (stats.Dmax_avg, stats.Dmax_std),
        "𝛂= %.1f±%.1f" % (stats.alpha_avg, stats.alpha_std),
        "S₀= %.4f±%.4f" % (stats.regularization_avg, stats.regularization_std),
        "χ²= %.2f±%.2f" % (stats.chi2r_avg, stats.chi2r_std),
        "logP= %.2f±%.2f" % (stats.evidence_avg, stats.evidence_std),
        "Rg= %.2f±%.2f" % (stats.Rg_avg, stats.Rg_std),
        "I₀= %.2f±%.2f" % (stats.I0_avg, stats.I0_std),
    ]
    with open(filename, "wt", encoding="utf-8") as out:
        out.write("# %s %s" % (source or filename, "\n"))
        for txt in res:
            out.write("# %s %s" % (txt, "\n"))
        out.write("%s# r\tp(r)\tsigma_p(r)%s" % ("\n", "\n"))
        for r, p, s in zip(
            stats.radius.astype(numpy.float32),
            stats.density_avg.astype(numpy.float32),
            stats.density_std.astype(numpy.float32),
        ):
            out.write("%s\t%s\t%s%s" % (r, p, s, "\n"))
    return filename + ": " + "; ".join(res)


StatsResult.save = save_bift

# Used in Cormap
GOF = namedtuple("GOF", ["n", "c", "P"])



# -*- coding: utf-8 -*-
#cython: boundscheck=False, wraparound=False, cdivision=True, embedsignature=True, language_level=3
# 
#    Project: freesas
#             https://github.com/kif/freesas
#
#    Copyright (C) 2017-2020  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from builtins import None, NotImplementedError

"""
Loosely based on the autoRg implementation in BioXTAS RAW by J. Hopkins
"""
__authors__ = ["Martha Brennich", "Jerome Kieffer"]
__license__ = "MIT"
__copyright__ = "2017 EMBL, 2020 ESRF"


class Error(Exception):
    """Base class for exceptions in this module."""
pass


class InsufficientDataError(Error):
    """Exception raised for errors in the input.
    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, msg="Not enough data do determine Rg"):
        self.expression = msg
        self.message = msg


class NoGuinierRegionError(Error):
    def __init__(self, qRg_max=None):
        self.expression = ""
        self.message = "No Guinier region found with reasonnable qRg < %s"%qRg_max


import cython
cimport numpy as cnumpy
import numpy as numpy 
from libc.math cimport sqrt, log, fabs, exp, atanh, ceil, NAN
from .isnan cimport isfinite 
from cython cimport floating
import logging
logger = logging.getLogger(__name__)
from .collections import RG_RESULT, FIT_RESULT

DTYPE = numpy.float64
ctypedef double DTYPE_t

# Definition of a few constants
cdef: 
    DTYPE_t[::1] WEIGHTS
    int RATIO_INTENSITY = 10   # start with range from Imax -> Imax/10
    DTYPE_t RG_MIN = 0.0098    # minimum acceptable value
    DTYPE_t Q_MIN_RG_MAX = 1.0 # maximum acceptable values
    DTYPE_t Q_MAX_RG_MAX = 1.3 # maximum acceptable values
    # DERIVED VALUES
    DTYPE_t RG2_MIN = RG_MIN*RG_MIN
    DTYPE_t Q2_MIN_RG2_MAX = Q_MIN_RG_MAX*Q_MIN_RG_MAX
    DTYPE_t Q2_MAX_RG2_MAX = Q_MAX_RG_MAX*Q_MAX_RG_MAX
    
qmaxrg_weight = 1.0
qminrg_weight = 0.1
rg_frac_err_weight = 1.0
i0_frac_err_weight = 1.0
r_sqr_weight = 1000
reduced_chi_sqr_weight = 0.0
window_size_weight = 6.0
    
_weights = numpy.array([qmaxrg_weight, qminrg_weight, rg_frac_err_weight, 
                        i0_frac_err_weight, r_sqr_weight, reduced_chi_sqr_weight, 
                        window_size_weight])
WEIGHTS = numpy.ascontiguousarray(_weights / _weights.sum(), DTYPE)


cdef inline DTYPE_t clamp(DTYPE_t x, 
                   DTYPE_t lower, 
                   DTYPE_t upper) nogil:
    "Equivalent of the numpy.clip"
    return min(upper, max(x, lower))


cdef int weighted_linear_fit(DTYPE_t[::1] datax, 
                             DTYPE_t[::1] datay, 
                             DTYPE_t[::1] weight, 
                             int data_start, 
                             int data_end, 
                             DTYPE_t[:, ::1] fit_mv, 
                             int position) nogil:
    """Calculates a fit to intercept-slope*x, weighted by w. s
        Input:
        :param x, y: two dataset to be fitted.
        :param w: The weight fot the individual points in x,y. Typically w would be 1/yerr**2 or 1/(yerr²+a²xerr²).
        :param data_start: first valid point in arrays
        :param data_end: last valid point in arrays
        :param fit_mv: output array with result: has to be an array of nx4
        :param position: the index in fit_mv where result should be written  
        
        Returns results: intercept, sigma_intercept, slope, sigma_slope
                        in the array fit_mv at the given position 
    """
    cdef: 
        int i, size
        DTYPE_t one_y, one_x, one_xy, one_xx, one_w, sigma_uxx, sigma_uxy, sigma_uyy
        DTYPE_t sigma_wy, sigma_wx, sigma_wxx, sigma_wxy, sigma_w, sigma_ux, sigma_uy
        DTYPE_t intercept, slope, xmean, ymean, ssxx, ssxy, ssyy, s, sigma_intercept, sigma_slope, xmean2
        DTYPE_t detA, reduced_chi_sqr

    intercept = slope = xmean = ymean = ssxx = ssyy = ssxy = s = sigma_intercept = sigma_slope = 0.0 
    sigma_uxx = sigma_wy = sigma_wx = sigma_wxy = sigma_wxx = xmean2 = sigma_w = 0.0 
    sigma_uyy = sigma_uy = sigma_ux = sigma_uxy = 0.0

# There should be a python function performing the sanitization
#     assert fit_mv.shape[0] >= position, "There is enough room for storing results"
#     assert fit_mv.shape[1] >= 8, "There is enough room for storing results"
    size = data_end - data_start
#     assert size > 2, "data range size should be >2"

    for i in range(data_start, data_end):
        one_w = weight[i]
        one_y = datay[i]
        one_x = datax[i]
        sigma_w += one_w
        sigma_wy += one_w * one_y
        sigma_wx += one_w * one_x
        sigma_wxy += one_w * one_x * one_y
        sigma_wxx += one_w * one_x * one_x
        sigma_uxx += one_x * one_x
        sigma_uxy += one_x * one_y
        sigma_uy += one_y
        sigma_ux += one_x
        sigma_uyy += one_y * one_y
    
    detA = sigma_wx * sigma_wx - sigma_w * sigma_wxx

    if fabs(detA) > 1e-100:
        # x = [-sigma_wxx*sigma_wy + sigma_wx*sigma_wxy,-sigma_wx*sigma_wy+sigma_w*sigma_wxy]/detA
        intercept = (-sigma_wxx * sigma_wy + sigma_wx * sigma_wxy) / detA
        slope = (+sigma_wx * sigma_wy - sigma_w * sigma_wxy) / detA
        xmean = sigma_ux / size
        ymean = sigma_uy / size
        xmean2 = xmean * xmean
        ssxx = sigma_uxx - size * xmean2 # *xmean
        ssyy = sigma_uyy - size * ymean * ymean
        s = sqrt((ssyy + slope * slope * ssxx) / (size - 2))
        sigma_intercept = sqrt(s * (1.0 / size + xmean2 / ssxx))
        sigma_slope = sqrt(s / ssxx)
        # xopt = (intercept, slope)
        # dx = (sigma_intercept, sigma_slope)
        # Returns results: intercept, sigma_intercept, slope, sigma_slope
        fit_mv[position, 0] = slope
        fit_mv[position, 1] = sigma_slope
        fit_mv[position, 2] = intercept
        fit_mv[position, 3] = sigma_intercept
        return 0
    else:
        return -1

cdef DTYPE_t calc_chi(DTYPE_t[::1] x, 
                      DTYPE_t[::1] y, 
                      DTYPE_t[::1] w,
                      int start, 
                      int end, 
                      DTYPE_t intercept, 
                      DTYPE_t slope,
                      DTYPE_t[:, ::1] fit_mv, 
                      int position) nogil:
    """Calculate the rvalue, r², chi² and weigted RMSD to be saved in fit_data"""
    cdef: 
        int idx, size
        DTYPE_t residual, sum_n, sum_y, sum_d, one_y, R2, mean_y, residual2, sum_w
        DTYPE_t chi2, one_w, pred_y, sum_dev2
    
    size = end - start
    sum_n = 0.0
    sum_y = 0.0
    sum_d = 0.0
    sum_w = 0.0
    chi2 = 0.0
    sum_dev2 = 0.0
    for idx in range(start, end):
        one_y = y[idx]
        one_w = w[idx]
        pred_y = slope * x[idx] + intercept 
        residual = one_y - pred_y
        residual2 = residual * residual
        sum_n += residual2
        sum_y += one_y
        sum_w += one_w  
        chi2 += residual2 * one_w
        sum_dev2 += residual2 * one_w / (sum_y * sum_y)
    mean_y = sum_y / size
    for idx in range(start, end):
        one_y = y[idx]
        residual = one_y - mean_y
        sum_d += residual * residual
    R2 = 1.0 - sum_n / sum_d
    
    fit_mv[position, 0] = sqrt(R2)               #r_value
    fit_mv[position, 1] = R2                     #r²_value
    fit_mv[position, 2] = chi2                   # chi²
    fit_mv[position, 3] = sqrt(sum_dev2/sum_w)   # weigted RMSD
    
    return R2

cdef inline void guinier_space(int start, 
                               int stop, 
                               DTYPE_t[::1] q, 
                               DTYPE_t[::1] intensity, 
                               DTYPE_t[::1] sigma, 
                               DTYPE_t[::1] q2, 
                               DTYPE_t[::1] lnI, 
                               DTYPE_t[::1] I2_over_sigma2) nogil:
        "Initialize q², ln(I) and I/sigma array"
        cdef:
            int idx
            DTYPE_t one_q, one_i, one_sigma
    
        q2[:] = 0.0
        lnI[:] = 0.0
        I2_over_sigma2[:] = 0.0
        
        for idx in range(start, stop):
            # populate the arrays for the fitting
            one_i = intensity[idx] 
            one_q = q[idx]
            q2[idx] = one_q * one_q
            lnI[idx] = log(one_i)
            one_sigma = sigma[idx]
            I2_over_sigma2[idx] = (one_i/one_sigma)**2


def linear_fit(x, y, w, 
               int start=0, 
               int stop=0):
    """Wrapper for testing of weighted_linear_fit from Python
    
    :param x, y: The dataset to be fitted.
    :param w: The weight fot the individual points in x,y. Typically w would be 1/yerr² or 1/(yerr²+a²xerr²)
    :param start: start position in the array x, y, and w
    :param stop: end position in the array x, y, and w (end position excluded 
    :return: FIT_RESULT namedtuple with slope, sigma_slope intercept, sigma_intercept 
    """
    cdef:
        DTYPE_t[::1] datax, datay, weight 
        int size, er
        DTYPE_t[:, ::1] fit_mv
    datax = numpy.ascontiguousarray(x, dtype=DTYPE) 
    datay = numpy.ascontiguousarray(y, dtype=DTYPE)
    weight= numpy.ascontiguousarray(w, dtype=DTYPE)
    size = datax.shape[0]
    assert datay.shape[0] == size, "y size matches"
    assert weight.shape[0] == size
    fit_mv = numpy.zeros((1, 8), dtype=DTYPE)
    if stop<=0:
        stop = size + stop
    with nogil:
        er = weighted_linear_fit(datax, datay, weight, start, size, fit_mv[:, :4], 0)
        if er != 0:
            with gil:
                raise ArithmeticError("Null determinant in linear regression") 
        else:
            calc_chi(datax, datay, weight,
                     start, stop, fit_mv[0, 2], fit_mv[0, 0], 
                     fit_mv[:, 4:], 0)
    return FIT_RESULT(*fit_mv[0])    


cdef class AutoGuinier:
    "Calculate the radius of gyration based on Guinier's formula. This class holds all constants"
    cdef:
        readonly int min_size, storage_size
        readonly DTYPE_t Rg_min, qminrgmax, qmaxrgmax, relax, aggregation_threshold
    
    def __cinit__(self, 
                  int min_size=3, 
                  DTYPE_t Rg_min=1.0,
                  DTYPE_t qmin_rgmax=1.0,
                  DTYPE_t qmax_rgmax=1.3,
                  DTYPE_t relax=1.2,
                  DTYPE_t aggregation_threshold=0.1):
        
        """Constructor of the class with:
        
        :param min_size: minimal size for the Guinier region in number of points (3, so that one can fit a parabola)
        :param Rg_min: minimum acceptable value for the radius of gyration (1nm)
        :param qmin_Rgmax: maximum acceptable value for the begining of Guinier region.
        :param qmax_Rgmax: maximum acceptable value for the end of Guinier region.
        :param relax: relax the qmax_rgmax constrain by this value for reduced quality data      
        #weights: those parameters are used to measure the best region
        #:param 
        """
        self.storage_size = 20
        self.min_size = min_size
        self.Rg_min = Rg_min
        self.qminrgmax = qmin_rgmax
        self.qmaxrgmax = qmax_rgmax
        self.relax = relax
        self.aggregation_threshold = aggregation_threshold
    
    def __dealloc__(self):
        self.storage_size = 0
        self.min_size = 0
        self.Rg_min = 0.0
        self.qminrgmax = 0
        self.qmaxrgmax = 0
        self.relax = 0
        self.aggregation_threshold = 0
        
    @cython.profile(True)
    cpdef curate_data(self,
                       data, 
                       DTYPE_t[::1] q, 
                       DTYPE_t[::1] intensity,
                       DTYPE_t[::1] sigma,
                       DTYPE_t Rg_min=-1.0,
                       DTYPE_t qRg_max=-1.0,
                       DTYPE_t relax=-1.0):
        """Clean up the input (data, 2D array of q, i, sigma)
        
        It removed negatives q, intensity, sigmas and also NaNs and infinites
        q, intensity and sigma are ouput array. 
                
        :param data: input data, array of (q, I, sigma) of size N
        :param q: output array, to be filled with valid values
        :param intensity: output array, to be filled with valid values
        :param sigma: output array, to be filled with valid values
        :return: Begining and end of the valid region
        
        Nota: May work with more then 3 col and discard subsequent columns in data
        """
        cdef:
            int idx, size, start, end
            DTYPE_t one_q, one_i, one_sigma, i_max, i_min

        size = data.shape[0]
        if qRg_max < 0.0:
            qRg_max = self.qmaxrgmax
        #qRg_max is now a hard threshold.
        if relax < 0.0:
            qRg_max *= self.relax
        else:
            qRg_max *= relax
        
        if Rg_min < 0.0:
            Rg_min = self.Rg_min
        
        q[:] = 0.0
        intensity[:] = 0.0
        sigma[:] = 0.0 
        # First pass: copy data and search for the max intensity
        start = self.min_size - 1 
        i_max = 0.0
        i_min = 0.0
        for idx in range(size):
            one_q = data[idx, 0]
            one_i = data[idx, 1]
            one_sigma = data[idx, 2]
            if isfinite(one_q) and one_q > 0.0 and \
               isfinite(one_i) and one_i > 0.0 and \
               isfinite(one_sigma) and one_sigma > 0.0:
                q[idx] = one_q
                intensity[idx] = one_i
                sigma[idx] = one_sigma
                if one_i > i_max:
                    i_min = i_max = one_i
                    start = idx
                if (idx-start)<=self.min_size:
                    i_min = min(i_min, one_i)
                if one_q*Rg_min < qRg_max:
                    end = idx
                else:
                    break
            else:
                q[idx] = NAN
                intensity[idx] = NAN
                sigma[idx] = NAN
                if (idx-start)<=self.min_size:
                    i_min = i_max = 0.0
                    start = - self.min_size -1
        return start, end+1

    def guinier_space(self, 
                       int start, 
                       int stop, 
                       DTYPE_t[::1] q, 
                       DTYPE_t[::1] intensity, 
                       DTYPE_t[::1] sigma, 
                       DTYPE_t[::1] q2, 
                       DTYPE_t[::1] lnI, 
                       DTYPE_t[::1] I2_over_sigma2):
        "Initialize q², ln(I) and I/sigma array"
        with nogil:
            guinier_space(start, stop, q, intensity, sigma, q2, lnI, I2_over_sigma2)

    @cython.profile(True)
    def many_fit(self, 
                 DTYPE_t[::1] q2_ary, 
                 DTYPE_t[::1] lnI_ary, 
                 DTYPE_t[::1] wg_ary,
                 int start,
                 int stop,  
                 DTYPE_t Rg_min=-1.0, 
                 DTYPE_t qRg_max=-1.0, 
                 DTYPE_t relax=-1.0):
        """Perform the linear regression for all reasonnably possible Guinier regions  
        
        :param q2_ary: array with q-squared (output of guinier_space)
        :param guinier_space: array with the log of the intensity (output of guinier_space)
        :param wg_ary: array with the weights (I²_over_sigma²)
        :param start: first valid point
        :param stop: last reasonnable valid point.
        :param Rg_min: Minimum acceptable radius of gyration (by default, the one from the class)
        :param qRg_max: End on the Guinier region (by default, the one from the class)
        :param relax: relax the qRg_max constrain by this value for reduced quality (by default, the one from the class)
        :return: an array of XXX-vector of floats, used as criteria to find the guinier region 
        
        0: Rg
        1: Rg_std
        2: I0
        3: I0_std
        4: start_point
        5: end_point
        6: quality
        7: agregated
        8: qRg_lower 
        9: qRg_upper
        10:14 slope, sigma_slope, intercept, sigma_intercept
        14:18 R, R², chi², RMDS
        18:20 q2Rg2_lower, q2Rg2_upper
        ...
        """
        cdef:
            int nb_fit, array_size, s, e, err
            DTYPE_t[:, ::1] result, tmp_mv
            DTYPE_t slope, sigma_slope, intercept, sigma_intercept, q2Rg2_lower, q2Rg2_upper
            DTYPE_t Rg2, Rg, qRg_upper, I0
            bint debug
        
        debug = logger.level<=logging.DEBUG
        nb_fit = 0
        array_size = 4096/(sizeof(DTYPE_t)*self.storage_size) #This correspond to one 4k page 
        result = numpy.zeros((array_size, self.storage_size), dtype=DTYPE)        

        with nogil:
            if qRg_max<0.0:
                qRg_max = self.qmaxrgmax
            #qRg_max is now a hard threshold.
            if relax<0.0:
                qRg_max *= self.relax
            else:
                qRg_max *= relax
            
            if Rg_min<0.0:
                Rg_min = self.Rg_min

            for s in range(start, stop-self.min_size):
                for e in range(s+self.min_size, stop):
                    # The 8 first parameters corresponds to the RG_RESULT
                    #result[nb_fit, 0] = Rg
                    #result[nb_fit, 1] = Rg_std
                    #result[nb_fit, 2] = I0
                    #result[nb_fit, 3] = I0_std
                    result[nb_fit, 4] = s
                    result[nb_fit, 5] = e 
                    #result[nb_fit, 6] = quality
                    #result[nb_fit, 7] = coef of the second order
                    #result[nb_fit, 8] = qRg_lower 
                    #result[nb_fit, 9] = qRg_upper
                    
                    # Coef 10-14 correspond to the linear regression
                    err = weighted_linear_fit(q2_ary, lnI_ary, wg_ary, s, e, result[:, 10:14], nb_fit)
                    if (err != 0):
                        if debug:
                            with gil:
                                logger.debug("position (%i,%i): Null determinant in linear regression", s, e)
                        result[nb_fit, :] = 0.0
                        continue                    
                    
                    slope = result[nb_fit, 10]
                    sigma_slope = result[nb_fit, 11] 
                    intercept = result[nb_fit, 12]
                    sigma_intercept = result[nb_fit, 13]
                    if slope >= 0:
                        result[nb_fit, :] = 0.0
                        if debug:
                            with gil:
                                logger.debug("position (%i,%i): Negative Rg²", s, e)
                        continue
                    # Coef 14-18 correspond to the assessement of the linear regression quality
                    if calc_chi(q2_ary, lnI_ary, wg_ary, s, e, 
                             intercept, slope, result[:, 14:18], nb_fit) <=0:
                        if debug:
                            with gil:
                                logger.debug("error in deviation calculation (%i, %i)", s, e)
                        # R² <0
                        result[nb_fit, :] = 0.0
                        continue
                    # Extract physical parameters:
                    Rg2 = -3.0*slope
                    result[nb_fit, 0] = Rg = sqrt(Rg2)
                    result[nb_fit, 1] = 0.5*sqrt(-3.0/slope)*sigma_slope # Rg_std
                    result[nb_fit, 2] = I0 = exp(intercept)
                    result[nb_fit, 3] = I0 * sigma_intercept #I0_std
                    
                    result[nb_fit, 18] = q2Rg2_lower = q2_ary[s] * Rg2
                    result[nb_fit, 19] = q2Rg2_upper = q2_ary[e-1] * Rg2
                    result[nb_fit, 8] = sqrt( q2Rg2_lower ) # qRg_lower
                    result[nb_fit, 9] = qRg_upper = sqrt( q2Rg2_upper )

                    if (Rg<Rg_min) or (qRg_upper>qRg_max):
                        if debug:
                            with gil:
                                logger.debug("Invalid qRg range (%i, %i) Rg %s > %s, qRg %s<%s", e, s, Rg, Rg_min, qRg_upper, qRg_max)
                        result[nb_fit, :] = 0.0
                        continue                        


                    #Claculate the descriptor for the quality
                    #result[nb_fit, 20] = result[nb_fit, 17]                          # 0 fit_score:  RMDS, normed
                    #result[nb_fit, 21] = 1.0 - <double>(e-s)/<double>(stop-start)    # 1 window_size_score: 
                    #result[nb_fit, 22] = q2Rg2_upper/self.q2maxrg2max                # 2 qmaxrg_score =  #quadratic penalty for qmax_Rg > 1.3
                    #result[nb_fit, 23] = q2Rg2_lower/self.q2minrg2max                # 3 qminrg_score =  #quadratic penalty for qmin_Rg > 1.0
                    #result[nb_fit, 24] = self.rg2_min/Rg2                            # 4 Rg_min score
                    #result[nb_fit, 25] = Rg_std/Rg                                   # 5 rg_error_score = 1.0 - sigma_Rg/Rg
                    #result[nb_fit, 26] = I0_std/I0                                   # 6 I0_error_score = 1.0 - sigma_I0/I0
                    
                    nb_fit +=1
                    if nb_fit >= array_size:
                        array_size *= 2
                        with gil:
                            tmp_mv = numpy.zeros((array_size, self.storage_size), dtype=DTYPE)
                            tmp_mv[:nb_fit, :] = result[:, :]
                            result = tmp_mv
        return numpy.asarray(result[:nb_fit])
    
    @cython.profile(True)
    def count_valid(self,
                    DTYPE_t[:, ::1] fit_result,
                    DTYPE_t qRg_max=-1.0, 
                    DTYPE_t relax=-1.0):
        """Count the number of valid Guinier intervals considering the qRg limit.  
        If no region is found, the constrain may be relaxed.
        
        Also searches for the maximum slope (absolute) value.

        :param fit_result: output of quality_fit: all possible Guinier reguions already fitted 
        :param qRg_max: upper bound of the Guinier region (1.3 is the standard value)
        :param relax: relax the qRg_max by this amount if no region are found.
        :return: number of regions, relaxed, actual qRg_max used, maximum slope
        """ 
        cdef:
            int i, size, cnt
            bint relaxed=0
            DTYPE_t aslope, aslope_max, qRg

        
        size = fit_result.shape[0]
        assert fit_result.shape[1] == self.storage_size, "fit_result shape is correct"
        with nogil:
            if qRg_max<0.0:
                qRg_max = self.qmaxrgmax
    
            #Start with strict mode, qRg_max being a soft-threshold
            aslope_max = 0.0
            cnt = 0
            for i in range(size):
                qRg = fit_result[i, 9]
                if qRg < qRg_max:
                    cnt += 1
                    aslope = -fit_result[i, 10] # slope is always negative by construction
                    if aslope>aslope_max:
                        aslope_max = aslope
            
            if cnt == 0:
                # There are not Guinier region whith reasonable qRg, let's relax the constrains
                relaxed = 1            
                if relax < 0.0:
                    qRg_max *= self.relax
                else:
                    qRg_max *= relax
                #qRg_max is now a hard threshold. 
                #Start with strict mode, qRg_max being a soft-threshold
                aslope_max = 0.0
                cnt = 0
                for i in range(size):
                    qRg = fit_result[i, 9]
                    if qRg < qRg_max:
                        cnt += 1
                        aslope = -fit_result[i, 10] # slope is always negative by construction
                        if aslope>aslope_max:
                            aslope_max = aslope
        return cnt, relaxed, qRg_max, aslope_max

    @cython.profile(True)
    cpdef (int, int) find_region(self, 
                                 DTYPE_t[:, ::1] fits, 
                                 DTYPE_t qRg_max=-1):
        """This function tries to extract the boundaries of the Guinier region from all fits.
        
        Each `start` and `stop` position is evaluated independantly, based on the average score of all contributing
        regions. Each region contributes with a weight calculated by:   
        * (q_max·Rg - q_min·Rg)/qRg_max --> in favor of large ranges
        * 1 / RMSD                      --> in favor of good quality data 
        
        For each start and end point, the contribution of all ranges are averaged out (using histograms-like techniques)
        The best solution is the (start,stop) couple position with the maximum score.
        
        To ensure start<stop, the stop is searched first (often less noisy) and the start is searched before stop.   
        
        :param fits: an array with all fits. Fits with q_max·Rg>qRg_max are ignored 
        :param qRg_max: the upper limit for searching the Guinier region. 
        :return 2-tuple with start-stop.
        """
        cdef:
            int start, stop, end, size, lower, upper, i
            cnumpy.int32_t[::1] unweigted_start, unweigted_stop
            DTYPE_t[::1] weigted_start,weigted_stop
            DTYPE_t max_weight, weight, qmin_Rg, qmax_Rg, RMSD 
        end = <int> numpy.max(fits[:,5])
        unweigted_start = numpy.zeros(end+1, dtype=numpy.int32)
        unweigted_stop = numpy.zeros(end+1, dtype=numpy.int32)
        weigted_start = numpy.zeros(end+1, dtype=DTYPE)
        weigted_stop = numpy.zeros(end+1, dtype=DTYPE)
        size = fits.shape[0]
        assert fits.shape[1] == self.storage_size, "size of fits matches"
        with nogil:
            if qRg_max<0.0:
                qRg_max = self.qmaxrgmax
            for i in range(size):
                qmax_Rg = fits[i, 9] #end of Guinier zone
                if qmax_Rg>qRg_max:
                    continue
                lower = <int> (fits[i, 4]) #lower index of the Guinier 
                upper = <int> (fits[i, 5]) #upperlower index of the Guinier
                qmin_Rg = fits[i, 8] #qRg at begining of Guinier zone
                RMSD = fits[i, 17]
                weight = (qmax_Rg - qmin_Rg)/RMSD # Empirically definied ... fits pretty well
                unweigted_start[lower] += 1
                unweigted_stop[upper] += 1
                weigted_start[lower] += weight
                weigted_stop[upper] += weight
            stop = 0
            max_weight = 0.0
            for i in range(end+1):
                if unweigted_stop[i]>0:
                    weight = weigted_stop[i]/unweigted_stop[i]
                    if weight>max_weight:
                        max_weight = weight
                        stop = i
            start=0
            max_weight = 0.0     
            for i in range(stop-self.min_size):
                if unweigted_start[i]>0:
                    weight = weigted_start[i]/unweigted_start[i]
                    if weight>max_weight:
                        max_weight = weight
                        start = i

        return start, stop
     
    @cython.profile(True)
    def slope_distribution(self,
                           DTYPE_t[:, ::1] fit_result,
                           int npt=1000,
                           DTYPE_t resolution=0.01,
                           DTYPE_t qRg_max=-1.0):
        """Find the most likely Guinier Region with updated parameters
        
        This function uses a 1D heat map of the |slope| values (with their errors) to assess the most likely Guinier region.
        This heat map is built of the sum of all gaussian curves, weighted by the q²Rg² range extension.
        
        Once the most likely |slope| has been found, the guinier region providing the nearest |slope| can be selected    
        
        :param fit_result: output of quality_fit: all possible Guinier reguions already fitted 
        :param npt: size of the distribution 
        :param resolution: The step size of the |slope|, from 0 to max(|slope|). Finer values provide better precision
        :param qRg_max: maximum allowed value for the upper limit of the Guinier region
        :return: the distribution of slopes.
        
        This function is finally not used.
        """
        cdef:
            int i, j, size
            DTYPE_t qRg, x, y, weight, aslope, sigma
            DTYPE_t[::1] distribution

        size = fit_result.shape[0]
        assert fit_result.shape[1] == self.storage_size, "fit_result shape is correct"

        distribution = numpy.zeros(npt, dtype=DTYPE)
        with nogil:
            for i in range(size):
                qRg = fit_result[i, 9]
                if qRg < qRg_max:
                    aslope = -fit_result[i, 10] # slope is always negative by contruction
                    sigma = fit_result[i, 11]
                    weight = (fit_result[i, 19]-fit_result[i, 18])/qRg_max
                    for j in range(npt):
                        x = j*resolution
                        y = weight*exp(-(x-aslope)**2/(2.0*sigma**2))/sigma
                        distribution[j] += y
        return numpy.asarray(distribution)
    
    @cython.profile(True)
    def average_values(self,
                       DTYPE_t[:, ::1] fit_result,
                       int start,
                       int stop
                       ):
        """Average out Rg and I0 and propagate errors from all sub-regions present in the main Guinier region
        
        :param: fit_result: output of quality_fit: all possible Guinier reguions already fitted 
        :paran start: the first index of the Guinier region
        :paran stop: the last index of the Guinier region
        :return: Rg_avg, Rg_std, I0_avg, I0_std, number of valid regions 
        """
        cdef:
            int i, good, size
            DTYPE_t wi, wr, swi, swr, srw, siw, Rg_avg, Rg_std, I0_avg, I0_std
        
        size = fit_result.shape[0]
        assert fit_result.shape[1] == self.storage_size, "fit_result shape is correct"
        with nogil:
            swi = swr = srw = siw = 0.0
            good = 0
            for i in range(size):
                if fit_result[i, 4]>=start and fit_result[i, 5]<=stop:
                    good +=1 
                    wr = 1/fit_result[i, 1]**2 # sigma_Rg
                    wi = 1/fit_result[i, 3]**2 # sigma_I0
                    swr += wr
                    swi += wi
                    srw += wr * fit_result[i,0] # Rg
                    siw += wi * fit_result[i,2] # I0
            Rg_avg = srw/swr
            I0_avg = siw/swi 
            swi = swr = srw = siw = 0.0
            good = 0
            for i in range(size):
                if fit_result[i, 4]>=start and fit_result[i, 5]<=stop:
                    good +=1 
                    wr = 1/fit_result[i, 1]**2 # sigma_Rg
                    wi = 1/fit_result[i, 3]**2 # sigma_I0
                    swr += wr
                    swi += wi
                    srw += wr * (fit_result[i,0]-Rg_avg)**2 # Rg
                    siw += wi * (fit_result[i,2]-I0_avg)**2 # I0
            Rg_std = sqrt(srw/swr)
            I0_std = sqrt(siw/swi)
        return Rg_avg, Rg_std, I0_avg, I0_std, good

    @cython.profile(True)
    def check_aggregation(self,
                          DTYPE_t[::1] q2, 
                          DTYPE_t[::1] lnI, 
                          DTYPE_t[::1] I2_over_sigma2, 
                          int start=0,
                          int end=0,
                          Rg=None,
                          threshold=None):
        """
        This function analyzes the curvature of a parabola fitted to the Guinier region 
        to check if the data indicate presence of larger aggragates
        
        A clear upwards curvature indicates aggregation.
        For convienniance , the polyfit from numpy is used which use the weights not squarred.
        
        :param q2: scattering vector squared
        :param lnI: logarithm if the intensity
        :param I2_over_sigma2: weight squared 
        :param start: the begining of the scan zone, argmax(I) is a good option, as early as possible
        :param end: the end of the scan zone, i.e. the end of the Guinier region
        :param Rg: the radius of gyration determined previously
        :param threshold: the value above which data are considered aggregated 
        :return: True if the protein is likely to be aggregated 
        if the threshold is None, return the value of the curvature*Rg**(-4)
        """
        cdef:
            DTYPE_t[::1] weight, coefs
            DTYPE_t value
        if end == 0:
            end=len(q2)
        else:
            assert len(q2)>=end, "q2 size matches"
        assert len(lnI)>=end, "lnI size matches"
        assert len(I2_over_sigma2)>=end, "I2_over_sigma2 size matches"
        weight = numpy.sqrt(I2_over_sigma2[start: end])
        try:
            coefs = numpy.polyfit(q2[start: end], lnI[start: end], 2, w=weight)
        except Exception as err:
            logger.error("Unable to fit parabola, %s: %s", type(err), err)
            return True
        if Rg:
            value = coefs[0]/Rg**4
        else:
            value = coefs[0]/(3*coefs[1])**2
    
        if threshold is False:
            return clamp(value, -1.0, 1.0)
        elif  threshold is None:
            return value>self.aggregation_threshold
        else:
            return value>threshold
    
    
    @cython.profile(True)
    cpdef DTYPE_t calc_quality(self, 
                               DTYPE_t Rg_avg, 
                               DTYPE_t Rg_std,
                               DTYPE_t qmin,
                               DTYPE_t qmax,
                               DTYPE_t aggregation,
                               DTYPE_t qRgmax = -1.0
                               ):
        """This function rates  the quality of the data.
        
        Unlike in `J. Appl. Cryst. (2007). 40, s223–s228`
        There I found no weight for how many consistent intervals were found.  
        
        :param Rg_avg: the average Rg found
        :param Rg_std: the standard deviation of Rg
        :param qmin: Scatteing vector of first point of the Guinier region
        :param qmax: Scatteing vector of last point of the Guinier region
        
        :return: quality indicator between 0 and 1
        """
        cdef:
            DTYPE_t quality
            DTYPE_t weight_Rg_dev, fit_Rg_dev, weight_qRgmax, fit_qRgmax, 
            DTYPE_t weight_drop, fit_drop, weight_aggregation, fit_aggregation, 
        
        if qRgmax <= 0:
            qRgmax = self.qmaxrgmax
        
        #0.58573097, 0.3070653 , 0.1795624 , 0.14731994, 2.6198585 
        
        #Quality form the fit of Rg
        weight_Rg_dev = 0.58573097
        fit_Rg_dev = 1.0 - clamp(Rg_std/Rg_avg, 0.0, 1.0)
         
        #Quality from the qRg max criteria  
        weight_qRgmax = 0.3070653
        fit_qRgmax = 1.0 - clamp(qmax*Rg_avg/qRgmax, 0.0, 1.0)
        
        #Quality from number of dropped points
        weight_drop = 0.1795624
        fit_drop = 1.0 - clamp(qmin/qmax, 0.0, 1.0)
        
        #quality from agregation value
        scale_aggregation = 2.5 
        weight_aggregation = 0.14731994
        fit_aggregation = 1.0 - clamp(scale_aggregation*fabs(aggregation), 0.0, 1.0)
        
        quality = fit_aggregation*weight_aggregation + \
                  fit_drop*weight_drop + \
                  fit_qRgmax*weight_qRgmax + \
                  fit_Rg_dev*weight_Rg_dev
        #as the sum of coef >1, can be larger than 1
        return clamp(quality, 0.0, 1.0)
    

    def fit(self, sasm):
        """This function automatically calculates the radius of gyration and scattering intensity at zero angle
        from a given scattering profile. It roughly follows the method used by the autorg function in the atsas package
        
        :param sasm: An array of q, I(q), dI(q)
        :return: RG_RESULT named tuple with the result of the fit
        """
        raise NotImplementedError()
    
################################################################################
# Old implementation from Matha    
################################################################################
def curate_data(floating[:, :] data, 
                 DTYPE_t[::1] q, 
                 DTYPE_t[::1] intensity,
                 DTYPE_t[::1] sigma,
                 DTYPE_t[::1] q2,
                 DTYPE_t[::1] log_intensity,
                 DTYPE_t[::1] weights,
                 int[::1] offsets,
                 int[::1] data_range):
    """Clean up the input (data, 2D array of q, i, sigma)
    
    It removed negatives q, intensity, sigmas and also NaNs and infinites
    q, intensity and sigma are ouput array. 

    If the data has more than 3 points, if any of the first three point is negative, 
    all data preceding it will be removed.
    
    we need also x: q*q, y: log I and w: (err/i)**(-2) 

    
    :param data: input data, array of q,i,sigma of size N
    :param q: output array, to be filled with valid values
    :param intensity: output array, to be filled with valid values
    :param sigma: output array, to be filled with valid values
    :param offsets: output array, provide the index in the input array for an 
                   index in the output array
    :return: the number of valid points in the array n <=N
    """
    cdef:
        int idx_in, idx_out, size_in, size_out, start, end, idx, start_curation
        DTYPE_t one_q, one_i, one_sigma, i_max, i_thres, tmp
        
    size_in = data.shape[0]
    
    # it may work with more then 3 col and discard subsequent columns
    # assert data.shape[1] == 3, "data has 3 columns" 
    assert q.size >= size_in, "size of q_array is valid"
    assert intensity.size >= size_in, "size of intensity array is valid"
    assert sigma.size >= size_in, "size of sigma array is valid"
    assert q2.size >= size_in, "size of q2 array is valid"
    assert log_intensity.size >= size_in, "size of log_intensity array is valid"
    assert weights.size >= size_in, "size of weights array is valid" 
    assert offsets.size >= size_in, "size of offsets array is valid"
    assert data_range.size >= 3, "data range holds enough space"
    
    # For safety: memset the arrays
    q[:] = 0.0
    intensity[:] = 0.0
    sigma[:] = 0.0
    q2[:] = 0.0
    log_intensity[:] = 0.0
    weights[:] = 0.0
    offsets[:] = 0
    data_range[:] = 0 
    start = 0  
    idx_out = 0
    i_max = 0.0
    start_curation = 0

    if size_in > 3:
        for idx_in in range(3):
            if data[idx_in, 1] < 0:
                start_curation = idx_in


    for idx_in in range(start_curation, size_in):
        one_q = data[idx_in, 0]
        one_i = data[idx_in, 1]
        one_sigma = data[idx_in, 2]
        if isfinite(one_q) and one_q > 0.0 and \
           isfinite(one_i) and one_i > 0.0 and \
           isfinite(one_sigma) and one_sigma > 0.0:
            q[idx_out] = one_q
            intensity[idx_out] = one_i
            sigma[idx_out] = one_sigma
            offsets[idx_out] = idx_in
            if one_i > i_max:
                i_max = one_i
                start = idx_out
            idx_out += 1
            
    # Second pass: focus on the valid region and prepare the 3 other arrays
    
    end = idx_out
    if end > start + 2:
        i_thres = (i_max + data[offsets[start + 1], 1] + data[offsets[start + 2], 1]) / (3 * RATIO_INTENSITY)
    else:
        i_thres = i_max / (RATIO_INTENSITY)
    
    for idx in range(start, idx_out):
        one_i = intensity[idx] 
        if one_i < i_thres:
            end = idx
            break
        else:
            # populate the arrays for the fitting  
            one_q = q[idx]
            q2[idx] = one_q * one_q
            log_intensity[idx] = log(one_i)
            # w = (i/sigma)**2
            one_sigma = sigma[idx]
            tmp = one_i / one_sigma
            weights[idx] = tmp * tmp
        
    data_range[0] = start        
    data_range[1] = end
    data_range[2] = idx_out   

def autoRg(sasm):
    """This function automatically calculates the radius of gyration and scattering intensity at zero angle
    from a given scattering profile. It roughly follows the method used by the autorg function in the atsas package
    Input:
    sasm: An array of q, I(q), dI(q)
    """
    cdef:
        DTYPE_t quality, intercept, slope, sigma_slope, lower, upper, r_sqr
        bint aggregated = 0
        cnumpy.ndarray qualities
        DTYPE_t[::1] q_ary, i_ary, sigma_ary, lgi_ary, q2_ary, wg_ary, 
        DTYPE_t[::1] fit_data
        int[::1] offsets, data_range
        int raw_size, curated_size, data_start, data_end, data_step
        int min_window, max_window, window_size, window_step 
        int start, end, nb_fit, array_size, block_size=39 #page of 4k
        int idx_min, idx_max, idx, err
        DTYPE_t[:, ::1] fit_mv, tmp_mv
        cnumpy.ndarray[DTYPE_t, ndim=2] fit_array
        
    raw_size = len(sasm)
    q_ary = numpy.empty(raw_size, dtype=DTYPE)
    i_ary = numpy.empty(raw_size, dtype=DTYPE)
    sigma_ary = numpy.empty(raw_size, dtype=DTYPE)
    q2_ary = numpy.empty(raw_size, dtype=DTYPE)
    lgi_ary = numpy.empty(raw_size, dtype=DTYPE)
    wg_ary = numpy.empty(raw_size, dtype=DTYPE)
    offsets = numpy.empty(raw_size, dtype=numpy.int32)
    data_range = numpy.zeros(3, dtype=numpy.int32)
    
    curate_data(sasm, q_ary, i_ary, sigma_ary, q2_ary, lgi_ary, wg_ary, offsets, data_range)
    
    data_start, data_end, curated_size = data_range
    
    logger.debug("raw size: %s, curated size: %s start: %s end: %s", raw_size, curated_size, data_start, data_end)
   
    if (data_end - data_start) < 10:
        raise InsufficientDataError()
  
    # Pick a minimum fitting window size. 10 is consistent with atsas autorg.
    min_window = 10
    max_window = data_end - data_start

    # It is very time consuming to search every possible window size and every 
    # possible starting point.
    # Here we define a subset to search.
    window_step = max_window // 10
    data_step = max_window // 50

    if window_step == 0:
        window_step = 1
    if data_step == 0:
        data_step = 1

    array_size = block_size
    nb_fit = 0
    fit_mv = numpy.zeros((array_size, 14), dtype=DTYPE)
    # This function takes every window size in the window list, stepts it through 
    # the data range, and fits it to get the RG and I0. If basic conditions are 
    # met, qmin*RG<1 and qmax*RG<1.35, and RG>0.1,
    # We keep the fit.
    with nogil:
        #for window_size in range(min_window, max_window + 1 , window_step):
        for window_size from min_window <= window_size < max_window + 1 by window_step:
            #for start in range(data_start, data_end - window_size, data_step):
            for start from data_start <= start < data_end - window_size by data_step:
                end = start + window_size
                #logger.debug("Fitting: %s , %s ", start,end)
                fit_mv[nb_fit, 0] = start
                fit_mv[nb_fit, 1] = window_size 
                fit_mv[nb_fit, 2] = q_ary[start]
                fit_mv[nb_fit, 3] = q_ary[end - 1]

                err = weighted_linear_fit(q2_ary, lgi_ary, wg_ary, start, end, fit_mv[:,4:8], nb_fit)

                if (err<0):
                    with gil:
                        logger.debug("Null determiant")
                        fit_mv[nb_fit, :] = 0.0
                        continue
    
                slope = -fit_mv[nb_fit, 4] 
                sigma_slope = fit_mv[nb_fit, 5] 
                intercept = fit_mv[nb_fit, 6]
                lower = q2_ary[start] * slope
                upper = q2_ary[start + window_size - 1] * slope

                fit_mv[nb_fit, 8] = lower 
                fit_mv[nb_fit, 9] = upper
                
                # check the validity of the model with some physics
                # i. e qmin*RG<1 and qmax*RG<1.35, and RG>0.1,
#                 if (slope >= 1/3.0) and (lower <= 1**2/3) and (upper <= 1.35**2/3) \
#                         and (sigma_slope / slope <= 1):
                if (slope > 3e-5) and (lower < 0.33) and (upper < 0.6075) \
                        and (sigma_slope / slope <= 1):
                    r_sqr = calc_chi(q2_ary, lgi_ary, wg_ary, start, end, 
                                     intercept, -slope, fit_mv[:,10:14], nb_fit)
                    if r_sqr > .15:
                        nb_fit += 1
                        if nb_fit >= array_size:
                            array_size *= 2
                            with gil:
                                tmp_mv = numpy.zeros((array_size, 14), dtype=DTYPE)
                                tmp_mv[:nb_fit, :] = fit_mv[:, :]
                                fit_mv = tmp_mv
                    else:
#                         with gil:
#                             logger.debug("(%i, %i) Invalid R²=%s", start, end, r_sqr)
                        #reset data
                        fit_mv[nb_fit, :] = 0.0
                else:
#                     with gil:
#                         logger.debug("(%i, %i) Invalid Rg slope: %s, lower: %s, upper %s sigma: %s\n %s", start, end, slope>3e-5, lower>0.33, upper<0.6075, 
#                                      sigma_slope / slope<1, numpy.asarray(fit_mv[nb_fit]))
                    fit_mv[nb_fit, :] = 0.0
    

    logger.debug("Number of valid fits: %s ", nb_fit)
                    
    if nb_fit == 0:
        #Extreme cases: may need to relax the parameters.
        pass
    
    if nb_fit > 0:
        fit_array = numpy.asarray(fit_mv)[:nb_fit, :]
        #Now we evaluate the quality of the fits based both on fitting data and on other criteria.

        #all_scores = []
        qmaxrg_score = 1.0 - numpy.absolute((fit_array[:, 9]-0.56)/0.56)
        qminrg_score = 1.0 - fit_array[:, 8]
        rg_frac_err_score = 1.0 - fit_array[:, 5]/fit_array[:, 4]
        i0_frac_err_score = 1.0 - fit_array[:, 7]/fit_array[:, 6]
        r_sqr_score = fit_array[:, 11]**4
        reduced_chi_sqr_score = 1.0 / fit_array[:,12] #Not right
        window_size_score = fit_array[:, 1] / max_window #float dividion forced by fit_array dtype 
        scores = numpy.array([qmaxrg_score, qminrg_score, rg_frac_err_score, i0_frac_err_score, r_sqr_score,
                              reduced_chi_sqr_score, window_size_score])
        qualities = numpy.dot(WEIGHTS, scores)
        #I have picked an aribtrary threshold here. Not sure if 0.6 is a good qualities cutoff or not.
        if qualities.max() > 0:# 0.5:
            # idx = qualities.argmax()
            # rg = fit_array[idx,4]
            # rger1 = fit_array[idx,5]
            # i0 = fit_array[idx,6]
            # i0er = fit_array[idx,7]
            # idx_min = fit_array[idx,0]
            # idx_max = fit_array[idx,0]+fit_array[idx,1]

            # try:
            #     #This adds in uncertainty based on the standard deviation of values with high qualities scores
            #     #again, the range of the qualities score is fairly aribtrary. It should be refined against real
            #     #data at some point.
            #     rger2 = fit_array[:,4][qualities>qualities[idx]-.1].std()
            #     rger = rger1 + rger2
            # except:
            #     rger = rger1

            try:
                idx = qualities.argmax()
                #rg = fit_array[:,4][qualities>qualities[idx]-.1].mean()
                
                rg = sqrt(-3. * fit_array[idx, 4])
                dber = fit_array[:, 5][qualities > qualities[idx] - .1].std()
                rger = 0.5 * sqrt(3. / rg) * dber
                i0 = exp(fit_array[idx, 6])
                #i0 = fit_array[:,6][qualities>qualities[idx]-.1].mean()
                daer = fit_array[:, 7][qualities > qualities[idx] - .1].std()
                i0er = i0 * daer
                idx_min = int(fit_array[idx, 0])
                idx_max = int(fit_array[idx, 0] + fit_array[idx, 1] - 1.0)
                idx_min_corr = numpy.argmin(numpy.absolute(sasm[:, 0] - fit_array[idx, 3]))
                idx_max_corr = numpy.argmin(numpy.absolute(sasm[:, 0] - fit_array[idx, 4]))
            except:
                idx = qualities.argmax()
                rg = sqrt(-3. * fit_array[idx, 4])
                rger = 0.5 * sqrt(-3. / fit_array[idx, 4]) * fit_array[idx, 5]
                i0 = exp(fit_array[idx, 6])
                i0er = i0 * fit_array[idx, 7]
                idx_min = int(fit_array[idx, 0])
                idx_max = int(fit_array[idx, 0] + fit_array[idx, 1] - 1.0)
            quality = qualities[idx]
        else:
          
            rg = -1
            rger = -1
            i0 = -1
            i0er = -1
            idx_min = -1
            idx_max = -1
            quality = 0

    else:
       
        rg = -1
        rger = -1
        i0 = -1
        i0er = -1
        idx_min = -1
        idx_max = -1
        quality = 0
        all_scores = []

    # managed by offsets
    # idx_min = idx_min + data_start
    # idx_max = idx_max + data_start

    #We could add another function here, if not good quality fits are found, either reiterate through the
    #the data and refit with looser criteria, or accept lower scores, possibly with larger error bars.

    aggregated = guinier.check_aggregation(q2_ary, lgi_ary, wg_ary, data_start, offsets[idx_max], Rg=rg)
    return RG_RESULT(rg, rger, i0, i0er, offsets[idx_min], offsets[idx_max], quality, aggregated)

guinier = AutoGuinier() 





# -*- coding: utf-8 -*-
"""Functions for calculating the radius of gyration and forward scattering intensity."""

__authors__ = ["Jerome Kieffer"]
__license__ = "MIT"
__copyright__ = "2020, ESRF"
__date__ = "05/06/2020"

import logging
import numpy
from scipy.optimize import curve_fit
from ._autorg import (  # pylint: disable=E0401
    RG_RESULT,
    autoRg,
    AutoGuinier,
    linear_fit,
    FIT_RESULT,
    guinier,
    NoGuinierRegionError,
    DTYPE,
    InsufficientDataError,
)


logger = logging.getLogger(__name__)


def auto_gpa(data, Rg_min=1.0, qRg_max=1.3, qRg_min=0.5):
    """
    Uses the GPA theory to guess quickly Rg, the
    radius of gyration and I0, the forwards scattering

    The theory is described in `Guinier peak analysis for visual and automated
    inspection of small-angle X-ray scattering data`
    Christopher D. Putnam
    J. Appl. Cryst. (2016). 49, 1412–1419

    This fits sqrt(q²Rg²)*exp(-q²Rg²/3)*I0/Rg to the curve I*q = f(q²)

    The Guinier region goes arbitrary from 0.5 to 1.3 q·Rg
    qRg_min and qRg_max can be provided

    :param data: the raw data read from disc. Only q and I are used.
    :param Rg_min: the minimal accpetable value for the radius of gyration
    :param qRg_max: the default upper bound for the Guinier region.
    :param qRg_min: the default lower bound for the Guinier region.
    :return: autRg result with limited information
    """

    def curate_data(data):
        q = data.T[0]
        I = data.T[1]
        err = data.T[2]

        start0 = numpy.argmax(I)
        stop0 = numpy.where(q > qRg_max / Rg_min)[0][0]

        range0 = slice(start0, stop0)
        q = q[range0]
        I = I[range0]
        err = err[range0]

        q2 = q ** 2
        lnI = numpy.log(I)
        I2_over_sigma2 = err ** 2 / I ** 2

        y = I * q
        p1 = numpy.argmax(y)

        # Those are guess from the max position:
        Rg = (1.5 / q2[p1]) ** 0.5
        I0 = I[p1] * numpy.exp(q2[p1] * Rg ** 2 / 3.0)

        # Let's cut-down the guinier region from 0.5-1.3 in qRg
        try:
            start1 = numpy.where(q > qRg_min / Rg)[0][0]
        except IndexError:
            start1 = None
        try:
            stop1 = numpy.where(q > qRg_max / Rg)[0][0]
        except IndexError:
            stop1 = None
        range1 = slice(start1, stop1)

        q1 = q[range1]
        I1 = I[range1]

        return q1, I1, Rg, I0, q2, lnI, I2_over_sigma2, start0

    q1, I1, Rg, I0, q2, lnI, I2_over_sigma2, start0 = curate_data(data)
    if len(q1) < 3:
        reduced_data = numpy.delete(data, start0, axis=0)
        q1, I1, Rg, I0, q2, lnI, I2_over_sigma2, start0 = curate_data(
            reduced_data
        )

    x = q1 * q1
    y = I1 * q1

    f = (
        lambda x, Rg, I0: I0
        / Rg
        * numpy.sqrt(x * Rg * Rg)
        * numpy.exp(-x * Rg * Rg / 3.0)
    )
    res = curve_fit(f, x, y, [Rg, I0])
    logger.debug(
        "GPA upgrade Rg %s-> %s and I0 %s -> %s", Rg, res[0][0], I0, res[0][1]
    )
    Rg, I0 = res[0]
    sigma_Rg, sigma_I0 = numpy.sqrt(numpy.diag(res[1]))
    end = numpy.where(data.T[0] > qRg_max / Rg)[0][0]
    start = numpy.where(data.T[0] > qRg_min / Rg)[0][0]
    aggregation = guinier.check_aggregation(
        q2, lnI, I2_over_sigma2, 0, end - start0, Rg=Rg, threshold=False
    )
    quality = guinier.calc_quality(
        Rg, sigma_Rg, data.T[0, start], data.T[0, end], aggregation, qRg_max
    )
    return RG_RESULT(
        Rg, sigma_Rg, I0, sigma_I0, start, end, quality, aggregation
    )


def auto_guinier(data, Rg_min=1.0, qRg_max=1.3, relax=1.2):
    """
    Yet another implementation of the Guinier fit

    The idea:
    * extract the reasonable range
    * convert to the Guinier space (ln(I) = f(q²)
    * scan all possible intervall
    * keep any with qRg_max<1.3 (or 1.5 in relaxed mode)
    * select the begining and the end of the guinier region according to the contribution of two parameters:
      - (q_max·Rg - q_min·Rg)/qRg_max --> in favor of large ranges
      - 1 / RMSD                      --> in favor of good quality data
      For each start and end point, the contribution of all ranges are averaged out (using histograms)
      The best solution is the start/end position with the maximum average.
    * All ranges within this region are averaged out to measure Rg, I0 and more importantly their deviation.
    * The quality is still to be calculated
    * Aggergation is assessed according a second order polynom fit.

    :param data: 2D array with (q,I,err)
    :param Rg_min: minimum value for Rg
    :param qRg_max: upper bound of the Guinier region
    :param relax: relaxation factor for the upper bound
    :param resolution: step size of the slope histogram
    :return: autRg result
    """

    raw_size = data.shape[0]
    q_ary = numpy.empty(raw_size, dtype=DTYPE)
    i_ary = numpy.empty(raw_size, dtype=DTYPE)
    sigma_ary = numpy.empty(raw_size, dtype=DTYPE)
    q2_ary = numpy.empty(raw_size, dtype=DTYPE)
    lnI_ary = numpy.empty(raw_size, dtype=DTYPE)
    wg_ary = numpy.empty(raw_size, dtype=DTYPE)

    start0, stop0 = guinier.curate_data(
        data, q_ary, i_ary, sigma_ary, Rg_min, qRg_max, relax
    )
    if start0 < 0:
        raise InsufficientDataError(
            "Minimum region size is %s" % guinier.min_size
        )
    guinier.guinier_space(
        start0, stop0, q_ary, i_ary, sigma_ary, q2_ary, lnI_ary, wg_ary
    )

    fits = guinier.many_fit(
        q2_ary, lnI_ary, wg_ary, start0, stop0, Rg_min, qRg_max, relax
    )

    cnt, relaxed, qRg_max, aslope_max = guinier.count_valid(
        fits, qRg_max, relax
    )
    # valid_fits = fits[fits[:, 9] < qRg_max]
    if cnt == 0:
        raise NoGuinierRegionError(qRg_max)

    # select the Guinier region based on all fits:
    start, stop = guinier.find_region(fits, qRg_max)

    # Now average out the
    Rg_avg, Rg_std, I0_avg, I0_std, good = guinier.average_values(
        fits, start, stop
    )

    aggregated = guinier.check_aggregation(
        q2_ary, lnI_ary, wg_ary, start0, stop, Rg=Rg_avg, threshold=False
    )
    quality = guinier.calc_quality(
        Rg_avg, Rg_std, q_ary[start], q_ary[stop], aggregated, qRg_max
    )
    result = RG_RESULT(
        Rg_avg, Rg_std, I0_avg, I0_std, start, stop, quality, aggregated
    )
    return result



#!usr/bin/env python
# coding: utf-8

__author__ = "Jérôme Kieffer"
__license__ = "MIT"
__date__ = "15/01/2021"
__copyright__ = "2015-2021, ESRF"

import sys
import unittest
from .test_all import suite


def run_tests():
    """Run test complete test_suite"""
    mysuite = suite()
    runner = unittest.TextTestRunner()
    if not runner.run(mysuite).wasSuccessful():
        print("Test suite failed")
        return 1
    else:
        print("Test suite succeeded")
        return 0


run = run_tests

if __name__ == '__main__':
    sys.exit(run_tests())



#!/usr/bin/python
# coding: utf-8
from __future__ import print_function

__author__ = "Jerome"
__license__ = "MIT"
__copyright__ = "2017, ESRF"

import numpy
import unittest

import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("test_cormap")
from .. import cormap


class TestCormap(unittest.TestCase):

    def test_longest(self):
        size = 1000
        target = 50
        start = 100

        data = numpy.ones(size, dtype="float32")
        res = cormap.measure_longest(data)
        self.assertEqual(res, size, msg="computed size is correct: positive")

        data -= 2
        res = cormap.measure_longest(data)
        self.assertEqual(res, size, msg="computed size is correct: negative")

        data[:] = 0
        data[start: start + target] = 1.0
        res = cormap.measure_longest(data)
        self.assertEqual(res, target, msg="computed size is correct: positive/zero")
        data = numpy.zeros(size, dtype="float32")
        data[start: start + target] = -1.0
        res = cormap.measure_longest(data)
        self.assertEqual(res, target, msg="computed size is correct: negative/zero")
        data = numpy.fromfunction(lambda n:(-1) ** n, (size,))
        data[start: start + target] = 1.0
        res = cormap.measure_longest(data)
        self.assertEqual(res, target + 1, msg="computed size is correct: positive/alternating")
        data = numpy.fromfunction(lambda n:(-1) ** n, (size,))
        data[start: start + target] = -1.0
        res = cormap.measure_longest(data)
        self.assertEqual(res, target + 1, msg="computed size is correct: negative/alternating")

    def test_stats(self):
        self.assertEqual(cormap.LROH.A(10, 0), 1)
        self.assertEqual(cormap.LROH.A(10, 1), 144)
        self.assertEqual(cormap.LROH.A(10, 2), 504)
        self.assertEqual(cormap.LROH.A(10, 10), 1024)
        self.assertEqual(cormap.LROH.A(10, 11), 1024)

        self.assertEqual(cormap.LROH.A(0, 3), 1)
        self.assertEqual(cormap.LROH.A(1, 3), 2)
        self.assertEqual(cormap.LROH.A(2, 3), 4)
        self.assertEqual(cormap.LROH.A(3, 3), 8)
        self.assertEqual(cormap.LROH.A(4, 3), 15)
        self.assertEqual(cormap.LROH.A(5, 3), 29)
        self.assertEqual(cormap.LROH.A(6, 3), 56)
        self.assertEqual(cormap.LROH.A(7, 3), 108)
        self.assertEqual(cormap.LROH.A(8, 3), 208)

        self.assertAlmostEqual(cormap.LROH(200, 0), 1)
        self.assertAlmostEqual(cormap.LROH(200, 4), 0.97, 2)
        self.assertAlmostEqual(cormap.LROH(200, 5), 0.80, 2)
        self.assertAlmostEqual(cormap.LROH(200, 6), 0.54, 2)
        self.assertAlmostEqual(cormap.LROH(200, 7), 0.32, 2)
        self.assertAlmostEqual(cormap.LROH(200, 8), 0.17, 2)
        self.assertAlmostEqual(cormap.LROH(200, 9), 0.09, 2)
        self.assertAlmostEqual(cormap.LROH(200, 10), 0.05, 2)
        self.assertAlmostEqual(cormap.LROH(200, 11), 0.02, 2)


def suite():
    testSuite = unittest.TestSuite()
    testSuite.addTest(TestCormap("test_longest"))
    testSuite.addTest(TestCormap("test_longest"))
    testSuite.addTest(TestCormap("test_stats"))
    return testSuite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

__author__ = "Jérôme Kieffer"
__license__ = "MIT"
__copyright__ = "2015, ESRF"
__date__ = "16/12/2015"

import numpy
import unittest
from .utilstests import get_datafile
from ..model import SASModel
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("cdistance_test")


class TestDistance(unittest.TestCase):
    testfile1 = get_datafile("model-01.pdb")
    testfile2 = get_datafile("dammif-01.pdb")

    def test_invariants(self):
        m = SASModel()
        m.read(self.testfile1)
        f_np, r_np, d_np = m.calc_invariants(False)
        f_cy, r_cy, d_cy = m.calc_invariants(True)
        self.assertAlmostEqual(f_np, f_cy, 10, "fineness is the same %s!=%s" % (f_np, f_cy))
        self.assertAlmostEqual(r_np, r_cy, 10, "Rg is the same %s!=%s" % (r_np, r_cy))
        self.assertAlmostEqual(d_np, d_cy, 10, "Dmax is the same %s!=%s" % (d_np, d_cy))

    def test_distance(self):
        m = SASModel()
        n = SASModel()
        m.read(self.testfile1)
        n.read(self.testfile2)
        f_np = m.dist(n, m.atoms, n.atoms, False)
        f_cy = m.dist(n, m.atoms, n.atoms, True)
        self.assertAlmostEqual(f_np, f_cy, 10, "distance is the same %s!=%s" % (f_np, f_cy))

    def test_same(self):
        m = SASModel()
        n = SASModel()
        m.read(self.testfile1)
        n.read(self.testfile1)
        numpy.random.shuffle(n.atoms)
        f_np = m.dist(n, m.atoms, n.atoms, False)
        f_cy = m.dist(n, m.atoms, n.atoms, True)
        self.assertAlmostEqual(f_np, 0, 10, "NSD not nul with np")
        self.assertAlmostEqual(f_cy, 0, 10, "NSD not nul with cy")


def suite():
    testSuite = unittest.TestSuite()
    testSuite.addTest(TestDistance("test_invariants"))
    testSuite.addTest(TestDistance("test_distance"))
    testSuite.addTest(TestDistance("test_same"))
    return testSuite

if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



# -*- coding: utf-8 -*-
#cython: boundscheck=False, wraparound=False, cdivision=True, embedsignature=True

__author__ = "Jerome Kieffer"
__license__ = "MIT"
__copyright__ = "2017, ESRF"

from cython cimport floating
from libc.stdlib cimport abs
import numpy 


cpdef int measure_longest(floating[::1] ary):
    """Measure the longest region where ary>0 or ary<0
    
    :param ary: numpy array or list of floats
    :return: the longest region with same sign
    """
    cdef: 
        int last = 0
        int longest = 0
        int acc = 0
        int i, d, size
        floating v
    size = ary.size
    with nogil:
        for i in range(size):
            v = ary[i]
            if v > 0:
                d = 1
            elif v < 0:
                d = -1
            else:
                d = 0
            if d * last <= 0:
                if abs(acc) > longest:
                    longest = abs(acc)
                acc = d
            else:
                acc = acc + d
            last = d
        if abs(acc) > longest:
            longest = abs(acc) 
    return longest



# -*- coding: utf-8 -*-
#
#    Project: freesas
#             https://github.com/kif/freesas
#
#    Copyright (C) 2017  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__authors__ = ["J. Kieffer"]
__license__ = "MIT"
__date__ = "10/06/2020"

import logging
import unittest
from math import sqrt, pi

import numpy
from scipy.stats import linregress

from .utilstests import get_datafile
from ..autorg import (
    autoRg,
    RG_RESULT,
    linear_fit,
    auto_gpa,
    auto_guinier,
)
from .._autorg import curate_data  # pylint: disable=E0401
from ..invariants import calc_Rambo_Tainer
from .._bift import distribution_sphere  # pylint: disable=E0401

logger = logging.getLogger(__name__)


def create_synthetic_data(R0=4, I0=100):
    """Create idealized data for a sphere of radius R0=4 whose Rg should be 4*sqrt(3/5)"""
    npt = 1000
    Dmax = 2 * R0
    size = 5000
    r = numpy.linspace(0, Dmax, npt + 1)
    p = distribution_sphere(I0, Dmax, npt)
    q = numpy.linspace(0, 10, size)
    qr = numpy.outer(q, r / pi)
    T = (4 * pi * (r[-1] - r[0]) / npt) * numpy.sinc(qr)
    I = T.dot(p)
    err = numpy.sqrt(I)
    return numpy.vstack((q, I, err)).T[1:]


class TestAutoRg(unittest.TestCase):
    testfile = get_datafile("bsa_005_sub.dat")

    def __init__(self, testName, **extra_kwargs):
        super().__init__(testName)
        self.extra_arg = extra_kwargs

    # Reference implementation
    atsas_autorg = {
        "Version": "Atsas 2.6.1",
        "Rg": 2.98016,
        "sigma_Rg": 0.156859,
        "I0": 61.3093,
        "sigma_I0": 0.0606315,
        "start_point": 46,
        "end_point": 95,
        "quality": 0.752564,
        "aggregated": 0,
    }

    def test_atsas(self):
        logger.info("test file: %s", self.testfile)
        data = numpy.loadtxt(self.testfile)
        atsas_result = self.atsas_autorg.copy()
        logger.debug("Reference version: %s" % atsas_result.pop("Version"))
        atsas_result = RG_RESULT(**atsas_result)
        free_result = autoRg(data)
        logger.debug("Ref: %s" % (atsas_result,))
        logger.debug("Obt: %s" % (free_result,))
        self.assertAlmostEqual(
            atsas_result.Rg, free_result.Rg, 1, "RG fits within 2 digits"
        )
        self.assertAlmostEqual(
            atsas_result.I0,
            free_result.I0,
            msg="I0 fits within +/- 1 ",
            delta=1,
        )
        self.assertAlmostEqual(
            atsas_result.quality,
            free_result.quality,
            0,
            msg="quality fits within 0 digits",
        )

    def test_synthetic(self):
        """Test based on sythetic data: a sphere of radius R0=4 which Rg should be 4*sqrt(3/5)"""
        R0 = 4
        I0 = 100
        data = create_synthetic_data(R0=R0, I0=I0)
        Rg = autoRg(data)
        logger.info("auto_rg %s", Rg)
        self.assertAlmostEqual(
            R0 * sqrt(3 / 5), Rg.Rg, 0, "Rg matches for a sphere"
        )
        self.assertGreater(
            R0 * sqrt(3 / 5),
            Rg.Rg - Rg.sigma_Rg,
            "Rg in range matches for a sphere",
        )
        self.assertLess(
            R0 * sqrt(3 / 5),
            Rg.Rg + Rg.sigma_Rg,
            "Rg in range matches for a sphere",
        )
        self.assertAlmostEqual(I0, Rg.I0, 0, "I0 matches for a sphere")
        self.assertGreater(I0, Rg.I0 - Rg.sigma_I0, "I0 matches for a sphere")
        self.assertLess(I0, Rg.I0 + Rg.sigma_I0, "I0 matches for a sphere")

        gpa = auto_gpa(data)
        logger.info("auto_gpa %s", gpa)
        self.assertAlmostEqual(
            gpa.Rg / (R0 * sqrt(3.0 / 5)), 1.00, 0, "Rg matches for a sphere"
        )
        self.assertAlmostEqual(gpa.I0 / I0, 1.0, 1, "I0 matches for a sphere")

        guinier = auto_guinier(data)
        logger.info("auto_guinier %s", guinier)
        self.assertAlmostEqual(
            R0 * sqrt(3.0 / 5), guinier.Rg, 0, "Rg matches for a sphere"
        )
        sigma_Rg = max(guinier.sigma_Rg, 1e-4)
        sigma_I0 = max(guinier.sigma_I0, 1e-4)
        self.assertGreater(
            R0 * sqrt(3.0 / 5),
            guinier.Rg - sigma_Rg,
            "Rg in range matches for a sphere",
        )
        self.assertLess(
            R0 * sqrt(3.0 / 5),
            guinier.Rg + sigma_Rg,
            "Rg in range matches for a sphere",
        )
        self.assertAlmostEqual(I0, guinier.I0, 0, "I0 matches for a sphere")
        self.assertGreater(
            I0, guinier.I0 - sigma_I0, "I0 matches for a sphere"
        )
        self.assertLess(I0, guinier.I0 + sigma_I0, "I0 matches for a sphere")

        # Check RT invarients...
        rt = calc_Rambo_Tainer(data, guinier)
        self.assertIsNotNone(
            rt, "Rambo-Tainer invariants are actually calculated"
        )

    def test_auto_gpa_with_outlier(self):

        """
        Test that auto_gpa gives reasonalbe results
        even if one data point is excessively large (e.g. hot pixel)
        """
        outlier_position = self.extra_arg["outlier_position"]
        R0 = 4
        I0 = 100
        data = create_synthetic_data(R0=R0, I0=I0)
        data[outlier_position, 1] *= 1000
        gpa = auto_gpa(data)
        logger.info("auto_gpa %s", gpa)
        self.assertAlmostEqual(
            gpa.Rg / (R0 * sqrt(3.0 / 5)),
            1.00,
            0,
            f"In case of outlier at {outlier_position} Rg matches for a sphere",
        )
        self.assertAlmostEqual(
            gpa.I0 / I0,
            1.0,
            1,
            f"In case of outlier at {outlier_position} I0 matches for a sphere",
        )


class TestFit(unittest.TestCase):
    # Testcase originally comes from wikipedia article on linear regression, expected results from scipy.stats.linregress

    def test_linear_fit_static(self):
        testx = [
            1.47,
            1.5,
            1.52,
            1.55,
            1.57,
            1.6,
            1.63,
            1.65,
            1.68,
            1.7,
            1.73,
            1.75,
            1.78,
            1.80,
            1.83,
        ]
        testy = [
            52.21,
            53.12,
            54.48,
            55.84,
            57.20,
            58.57,
            59.93,
            61.29,
            63.11,
            64.47,
            66.28,
            68.1,
            69.92,
            72.19,
            74.46,
        ]
        testw = [1.0] * 15
        testintercept = -39.061956
        testslope = +61.2721865
        fit_result = linear_fit(testx, testy, testw)
        # print(fit_result)
        self.assertAlmostEqual(
            fit_result.intercept,
            testintercept,
            5,
            "Intercept fits wihtin 4(?) digits",
        )
        self.assertAlmostEqual(
            fit_result.slope, testslope, 5, "Intercept fits wihtin 4(?) digits"
        )

    def test_linspace(self):
        size = 100
        x = numpy.linspace(-10, 10, size)
        y = numpy.linspace(10, 0, size)
        w = numpy.random.random(size)
        fit_result = linear_fit(x, y, w)
        # print(fit_result)
        self.assertAlmostEqual(
            fit_result.intercept, 5, 5, "Intercept fits wihtin 4(?) digits"
        )
        self.assertAlmostEqual(
            fit_result.slope, -0.5, 5, "Intercept fits wihtin 4(?) digits"
        )

    def test_random(self):

        """
        Tests that our linear regression implementation
        gives the same results as scipy.stats for random data
        """
        size = 100
        x = numpy.random.random(size)
        y = 1.6 * x + 5 + numpy.random.random(size)
        w = numpy.ones(size)
        fit_result = linear_fit(x, y, w)
        ref = linregress(x, y)
        self.assertAlmostEqual(
            fit_result.intercept,
            ref[1],
            5,
            "Intercept fits wihtin 4(?) digits",
        )
        self.assertAlmostEqual(
            fit_result.slope, ref[0], 5, "Intercept fits wihtin 4(?) digits"
        )
        self.assertAlmostEqual(
            fit_result.R2,
            ref.rvalue ** 2,
            5,
            "R² value matcheswihtin 4(?) digits",
        )


class TestDataCuration(unittest.TestCase):
    """Tests for the curate_data function."""

    testfile = get_datafile("bsa_005_sub.dat")

    def __init__(self, testName, **extra_kwargs):
        super().__init__(testName)
        self.extra_arg = extra_kwargs

    def test_curate_data_BM29_bsa(self):
        """Test data curration of "nice" BM29 data."""
        logger.info("test file: %s", self.testfile)
        data = numpy.loadtxt(self.testfile)
        DTYPE = numpy.float64
        raw_size = len(data)
        q_ary = numpy.empty(raw_size, dtype=DTYPE)
        i_ary = numpy.empty(raw_size, dtype=DTYPE)
        sigma_ary = numpy.empty(raw_size, dtype=DTYPE)
        q2_ary = numpy.empty(raw_size, dtype=DTYPE)
        lgi_ary = numpy.empty(raw_size, dtype=DTYPE)
        wg_ary = numpy.empty(raw_size, dtype=DTYPE)
        offsets = numpy.empty(raw_size, dtype=numpy.int32)
        data_range = numpy.zeros(3, dtype=numpy.int32)

        curate_data(
            data,
            q_ary,
            i_ary,
            sigma_ary,
            q2_ary,
            lgi_ary,
            wg_ary,
            offsets,
            data_range,
        )

        self.assertListEqual(
            list(data_range),
            [2, 203, 1033],
            msg="reproduced expected BM29 data range",
        )

    def test_curate_synthetic_data(self):
        """Test that for idealized data the cut-off is at i0/10."""
        data = create_synthetic_data()
        I_one = data[0, 1]
        DTYPE = numpy.float64
        raw_size = len(data)
        q_ary = numpy.empty(raw_size, dtype=DTYPE)
        i_ary = numpy.empty(raw_size, dtype=DTYPE)
        sigma_ary = numpy.empty(raw_size, dtype=DTYPE)
        q2_ary = numpy.empty(raw_size, dtype=DTYPE)
        lgi_ary = numpy.empty(raw_size, dtype=DTYPE)
        wg_ary = numpy.empty(raw_size, dtype=DTYPE)
        offsets = numpy.empty(raw_size, dtype=numpy.int32)
        data_range = numpy.zeros(3, dtype=numpy.int32)

        curate_data(
            data,
            q_ary,
            i_ary,
            sigma_ary,
            q2_ary,
            lgi_ary,
            wg_ary,
            offsets,
            data_range,
        )

        self.assertEqual(
            offsets[0],
            0,
            msg="curated data for artificial data starts at 0",
        )

        self.assertTrue(
            data[data_range[1] - 1, 1] > I_one / 10
            and data[data_range[1] + 1, 1] < I_one / 10,
            msg="curated data for artificial data ends at approx. I0/10",
        )

    def test_curate_synthetic_data_with_negative_points(self):
        """Test that if one of the first three points is negative, all date before it gets ignored."""
        negative_point_index = self.extra_arg["negative_point_index"]

        data = create_synthetic_data()
        DTYPE = numpy.float64
        raw_size = len(data)
        data[negative_point_index, 1] = -1

        q_ary = numpy.empty(raw_size, dtype=DTYPE)
        i_ary = numpy.empty(raw_size, dtype=DTYPE)
        sigma_ary = numpy.empty(raw_size, dtype=DTYPE)
        q2_ary = numpy.empty(raw_size, dtype=DTYPE)
        lgi_ary = numpy.empty(raw_size, dtype=DTYPE)
        wg_ary = numpy.empty(raw_size, dtype=DTYPE)
        offsets = numpy.empty(raw_size, dtype=numpy.int32)
        data_range = numpy.zeros(3, dtype=numpy.int32)

        curate_data(
            data,
            q_ary,
            i_ary,
            sigma_ary,
            q2_ary,
            lgi_ary,
            wg_ary,
            offsets,
            data_range,
        )

        self.assertEqual(
            offsets[0],
            negative_point_index + 1,
            msg=f"curated data for artificial data starts after negative data point for negative point at {negative_point_index + 1}",
        )

        self.assertTrue(
            data[offsets[data_range[1]] - 1, 1]
            > data[negative_point_index + 1, 1] / 10
            and data[offsets[data_range[1]] + 1, 1]
            < data[negative_point_index + 1, 1] / 10,
            msg=f"curated data for artificial data ends at approx. I(point after negaitve point)/10 if negative point at {negative_point_index + 1}",
        )


def suite():
    """Generic builder for the test suite."""
    testSuite = unittest.TestSuite()
    testSuite.addTest(TestAutoRg("test_atsas"))
    testSuite.addTest(TestAutoRg("test_synthetic"))
    for outlier_position in range(3):
        testSuite.addTest(
            TestAutoRg(
                "test_auto_gpa_with_outlier", outlier_position=outlier_position
            )
        )
    testSuite.addTest(TestFit("test_linear_fit_static"))
    testSuite.addTest(TestFit("test_linspace"))
    testSuite.addTest(TestDataCuration("test_curate_data_BM29_bsa"))
    testSuite.addTest(TestDataCuration("test_curate_synthetic_data"))
    for negative_point_index in range(3):
        testSuite.addTest(
            TestDataCuration(
                "test_curate_synthetic_data_with_negative_points",
                negative_point_index=negative_point_index,
            )
        )
    return testSuite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



"""This module provides a function which reads in the data,
performs the guinier fit with a given algotithm and reates the input."""

__authors__ = ["Martha Brennich"]
__contact__ = "martha.brennich@googlemail.com"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "21/03/2021"
__status__ = "development"
__docformat__ = "restructuredtext"

import sys
import logging
import platform
from os import linesep as os_linesep
from pathlib import Path
from contextlib import contextmanager
from typing import Callable, List, Optional, IO, Generator
from numpy import ndarray
from .autorg import (
    RG_RESULT,
    InsufficientDataError,
    NoGuinierRegionError,
)
from .sasio import (
    load_scattering_data,
    convert_inverse_angstrom_to_nanometer,
)
from .sas_argparser import GuinierParser


def set_logging_level(verbose_flag: int) -> None:
    """
    Set logging level according to verbose flag of argparser
    :param verbose_flag: int flag for logging level
    """
    if verbose_flag == 1:
        logging.root.setLevel(logging.INFO)
    elif verbose_flag >= 2:
        logging.root.setLevel(logging.DEBUG)


def collect_files(file_list: List[str]) -> List[Path]:
    """
    Take file list from argparser and return list of paths
    :param file_list: file list as returned by the argparser
    :return: A list of Path objects which includes only existing files
    """
    files = [Path(i) for i in file_list if Path(i).exists()]
    if platform.system() == "Windows" and files == []:
        files = list(Path.cwd().glob(file_list[0]))
        files.sort()
    return files


@contextmanager
def get_output_destination(
    output_path: Optional[Path] = None,
) -> Generator[IO[str], None, None]:
    """
    Return file or stdout object to write output to
    :param output_path: None if output to stdout, else Path to outputfile
    :return: opened file with write access or sys.stdout
    """
    # pylint: disable=R1705
    if output_path is not None:
        with open(output_path, "w") as destination:
            yield destination
    else:
        yield sys.stdout


def get_linesep(output_destination: IO[str]) -> str:
    """
    Get the appropriate linesep depending on the output destination.
    :param output_destination: an IO object, e.g. an open file or stdout
    :return: string with the correct linesep
    """
    # pylint: disable=R1705
    if output_destination == sys.stdout:
        return os_linesep
    else:
        return "\n"


def get_header(linesep: str, output_format: Optional[str] = None) -> str:
    """Return appropriate header line for selected output format
    :param output_format: output format from string parser
    :param linesep: correct linesep for chosen destination
    :return: a one-line string"""
    # pylint: disable=R1705
    if output_format == "csv":
        return (
            ",".join(
                (
                    "File",
                    "Rg",
                    "Rg StDev",
                    "I(0)",
                    "I(0) StDev",
                    "First point",
                    "Last point",
                    "Quality,Aggregated",
                )
            )
            + linesep
        )
    else:
        return ""


def rg_result_to_output_line(
    rg_result: RG_RESULT,
    afile: Path,
    linesep: str,
    output_format: Optional[str] = None,
) -> str:
    """Return result line formatted according to selected output format
    :param rg_result: Result of an rg fit
    :param afile: The name of the file that was processed
    :param output_format: The chosen output format
    :param linesep: correct linesep for chosen destination
    :return: a one-line string including linesep"""
    # pylint: disable=R1705
    if output_format == "csv":
        return (
            ",".join(
                [
                    f"{afile}",
                    f"{rg_result.Rg:6.4f}",
                    f"{rg_result.sigma_Rg:6.4f}",
                    f"{rg_result.I0:6.4f}",
                    f"{rg_result.sigma_I0:6.4f}",
                    f"{rg_result.start_point:3}",
                    f"{rg_result.end_point:3}",
                    f"{rg_result.quality:6.4f}",
                    f"{rg_result.aggregated:6.4f}",
                ]
            )
            + linesep
        )
    elif output_format == "ssv":
        return (
            " ".join(
                [
                    f"{rg_result.Rg:6.4f}",
                    f"{rg_result.sigma_Rg:6.4f}",
                    f"{rg_result.I0:6.4f}",
                    f"{rg_result.sigma_I0:6.4f}",
                    f"{rg_result.start_point:3}",
                    f"{rg_result.end_point:3}",
                    f"{rg_result.quality:6.4f}",
                    f"{rg_result.aggregated:6.4f}",
                    f"{afile}",
                ]
            )
            + linesep
        )
    else:
        return f"{afile} {rg_result}{linesep}"


def run_guinier_fit(
    fit_function: Callable[[ndarray], RG_RESULT],
    parser: GuinierParser,
    logger: logging.Logger,
) -> None:
    """
    reads in the data, performs the guinier fit with a given algotithm and
    creates the
    :param fit_function : A Guinier fit function data -> RG_RESULT
    :param parser: a function that returns the output of argparse.parse()
    :param logger: a Logger
    """
    args = parser.parse_args()
    set_logging_level(args.verbose)
    files = collect_files(args.file)
    logger.debug("%s input files", len(files))

    with get_output_destination(args.output) as output_destination:
        linesep = get_linesep(output_destination)

        output_destination.write(
            get_header(
                linesep,
                args.format,
            )
        )

        for afile in files:
            logger.info("Processing %s", afile)
            try:
                data = load_scattering_data(afile)
            except OSError:
                logger.error("Unable to read file %s", afile)
            except ValueError:
                logger.error("Unable to parse file %s", afile)
            else:
                if args.unit == "Å":
                    data = convert_inverse_angstrom_to_nanometer(data)
                try:
                    rg_result = fit_function(data)
                except (
                    InsufficientDataError,
                    NoGuinierRegionError,
                    ValueError,
                    IndexError,
                ) as err:
                    sys.stderr.write(
                        f"{afile}, {err.__class__.__name__}: {err} {os_linesep}"
                    )
                else:
                    res = rg_result_to_output_line(
                        rg_result,
                        afile,
                        linesep,
                        args.format,
                    )
                    output_destination.write(res)
                    output_destination.flush()
